import traceback
from gempyp.libs.common import get_reason_of_failure
import numpy
import math
import itertools
import collections
from gempyp.libs.enums.status import status
import time


def df_compare(src_df, tgt_df, key, logger, reporter, configData):
    try:
        logger.info("In df_compare Function")
        headers = list(set(list(src_df.columns)) - set(key))
        src_df["key"] = src_df[key].apply(
            lambda row: "----".join(row.values.astype(str)), axis=1
        )
        tgt_df["key"] = tgt_df[key].apply(
            lambda row: "----".join(row.values.astype(str)), axis=1
        )
        src_df.set_index("key", inplace=True)
        tgt_df.set_index("key", inplace=True)
        src_df.drop(key, axis=1, inplace=True)
        tgt_df.drop(key, axis=1, inplace=True)
        src_key_values = src_df.index.values
        tgt_key_values = tgt_df.index.values
        common_keys = list(set(src_key_values) & set(tgt_key_values))
        keys_only_in_src = list(set(src_key_values) - set(tgt_key_values))
        keys_only_in_tgt = list(set(tgt_key_values) - set(src_key_values))
        src_df.drop(keys_only_in_src, inplace=True)
        tgt_df.drop(keys_only_in_tgt, inplace=True)
        # if 'SKIP_COLUMN' in configData:
        #     try:
            # skip_column = configData["SKIP_COLUMN"].split(',')
        
            # headers = list(set(headers)-set(skip_column))
        #     except Exception as e:
        #         reporter.addRow("While Skipping Columns",get_reason_of_failure(traceback.format_exc(), e),status.INFO)
        """calling src and tgt for getting different keys"""
        src_key_dict = addDiffKeysDict(keys_only_in_src, "Source", key, logger)
        tgt_key_dict = addDiffKeysDict(keys_only_in_tgt, "Target", key, logger)
        # for keys in src_key_dict.keys():
        #     src_key_dict[keys] = src_key_dict[keys] + tgt_key_dict[keys]
        diff_keys_dict = collections.defaultdict(list)
        for keys, val in itertools.chain(src_key_dict.items(), tgt_key_dict.items()):
            diff_keys_dict[keys] += val
        diff_keys_dict = dict(diff_keys_dict)
        src_df.index.astype(str)
        tgt_df.index.astype(str)
        src_df.sort_values(by=["key"], inplace=True)
        tgt_df.sort_values(by=["key"], inplace=True)
        common_keys.sort()
        value_dict = getValueDict(
            src_df, tgt_df, common_keys, headers, key, configData, logger, reporter
        )
        keys_length = {
            "keys_only_in_src": len(keys_only_in_src),
            "keys_only_in_tgt": len(keys_only_in_tgt),
            "common_keys": len(common_keys)
        }
        if "TOLERANCE" in configData:
            reporter.addMisc("TOLERANCE",str(configData.get("TOLERANCE",0)))
        if "ROUND_OFF" in configData:
            reporter.addMisc("ROUND OFF",str(configData.get("ROUND_OFF",0)))
        return value_dict, diff_keys_dict, keys_length
    except Exception as e:
        traceback.print_exc()
        reporter.addRow("Comparing data", "Error occurred", status.ERR)
        reporter.addMisc(
            "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e)
        )


def addDiffKeysDict(_list, db, keys, logger):
    logger.info("In addDiffKeysDict Function")
    key_dict = {}
    dict1 = {"Reason-of-Failure": []}
    li1 = []
    for i in keys:
        li1.append([])
    if _list:
        for i in _list:
            li = i.split("----")
            for i in range(len(li)):
                key = keys[i]
                li1[i].append(li[i])
                value = li1[i]
                key_dict[key] = value
            dict1.get("Reason-of-Failure").append(f"keys only in {db}")
    key_dict.update(dict1)
    return key_dict


def compareValues(commonList: list, src_df, tgt_df, headers, keys, configData, logger,cut_out,count):
    logger.info("In compareValues Function")
    dummy_dict = {
        "Column-Name": [],
        "Source-Value": [],
        "Target-Value": [],
        "Reason-of-Failure": [],
    }
    comm_dict = {}

    if commonList:
        for key_val in commonList:
            for field in headers:
                src_val = src_df.loc[key_val, field]
                tgt_val = tgt_df.loc[key_val, field]
                if src_val == src_val or tgt_val == tgt_val:
                    if "ROUND_OFF" in configData:
                        # self.reporter.addMisc("",str(self.configData["THRESHOLD"]))
                        if (
                            type(src_val) == numpy.float64
                            and math.isnan(src_val) == False
                        ):
                            src_val = truncate(src_val, int(configData["ROUND_OFF"]))
                        if (
                            type(tgt_val) == numpy.float64
                            and math.isnan(tgt_val) == False
                        ):
                            tgt_val = truncate(tgt_val, int(configData["ROUND_OFF"]))

                    if type(src_val) == numpy.float64 and type(tgt_val) == numpy.float64:
                        if math.isnan(src_val) == False and math.isnan(tgt_val) == False:
                            if abs(src_val - tgt_val) > float(configData.get("TOLERANCE",0)):
                                li = key_val.split("----")
                                #this is for getting the key value
                                for i in range(len(li)):
                                    key = keys[i]
                                    li1 = []
                                    li1.append(li[i])
                                    comm_dict[key] = comm_dict.get(key, []) + li1
                                dummy_dict.get("Column-Name").append(field)
                                dummy_dict.get("Source-Value").append(src_val)
                                dummy_dict.get("Target-Value").append(tgt_val)
                                dummy_dict.get("Reason-of-Failure").append(
                                    "Difference In Value"
                                    )   

                    elif src_val != tgt_val and type(src_val) == type(tgt_val):
                        li = key_val.split("----")
                        for i in range(len(li)):
                            key = keys[i]
                            li1 = []
                            li1.append(li[i])
                            comm_dict[key] = comm_dict.get(key, []) + li1
                        dummy_dict.get("Column-Name").append(field)
                        dummy_dict.get("Source-Value").append(src_val)
                        dummy_dict.get("Target-Value").append(tgt_val)
                        dummy_dict.get("Reason-of-Failure").append(
                            "Difference In Value"
                        )

                    elif type(src_val) != type(tgt_val):
                        li = key_val.split("----")
                        for i in range(len(li)):
                            key = keys[i]
                            li1 = []
                            li1.append(li[i])
                            comm_dict[key] = comm_dict.get(key, []) + li1
                        dummy_dict.get("Column-Name").append(field)
                        dummy_dict.get("Source-Value").append(src_val)
                        dummy_dict.get("Target-Value").append(tgt_val)
                        dummy_dict.get("Reason-of-Failure").append(
                            "Difference In Datatype"
                        )
            if count + len(dummy_dict['Reason-of-Failure']) > int(cut_out):
                break   
        comm_dict.update(dummy_dict)
        return comm_dict


def truncate(f, n):
    return math.floor(f * 10**n) / 10**n


def getValueDict(src_df, tgt_df, common_keys, headers, key, configData, logger, reporter):
    logger.info("In getValueDict Function")
    splitSize = 100000
    """this code is for deciding cutout range"""
    if len(common_keys) > 100000:
        length = (10*len(common_keys))//100
        cut_out = configData.get('CUT_OUT',length)
    else:
        cut_out = configData.get('CUT_OUT',100000)
        
    common_keys_splited = [
        common_keys[x : x + splitSize] for x in range(0, len(common_keys), splitSize)
    ]
    src_df_splited = [src_df[x : x + splitSize] for x in range(0, len(src_df), splitSize)]
    tgt_df_splited = [tgt_df[x : x + splitSize] for x in range(0, len(tgt_df), splitSize)]
    # final_value_diffs = collections.defaultdict(list)
    final_value_diffs = {
            "Column-Name": [],
            "Source-Value": [],
            "Target-Value": [],
            "Reason-of-Failure": [],
        }
    count = 0
    for i in range(len(common_keys_splited)):
        logger.info(time.time())
        # data = {"common_keys_splited":common_keys_splited[i],"src_df_splited":src_df_splited[i],"tgt_df_splited":tgt_df_splited[i],"headers":headers,"key":key,"configData":configData,"logger":logger,"cut_out":cut_out,"count":count}
        chunk_diffs = compareValues(
            common_keys_splited[i],
            src_df_splited[i],
            tgt_df_splited[i],
            headers,
            key,
            configData,
            logger,
            cut_out,
            count
        )
        for keys in chunk_diffs.keys():
            if keys in final_value_diffs.keys():
                final_value_diffs[keys] = final_value_diffs[keys] + chunk_diffs[keys]
            else:
                final_value_diffs[keys] = chunk_diffs[keys]
        # for key, val in itertools.chain(final_value_diffs.items(), chunk_diffs.items()):
        # final_value_diffs[key] += val
        # final_value_diffs = dict(final_value_diffs)
        count = len(final_value_diffs["Reason-of-Failure"])
        if count > int(cut_out):
            reporter.addRow("Stopping execution","Mismatch count is greater than {}".format(cut_out),status.INFO)
            break
        
        logger.info(time.time())
    return final_value_diffs



import requests
import os
import json
import logging
import tempfile
import uuid
from requests.auth import HTTPBasicAuth
from base64 import b64encode

def fetchFileFromGit(link,branch,username=None,bearer_token=None):  ################ post 1.0.4
        list=link.split("/")
        api = f'https://api.github.com/repos/{list[3]}/{list[4]}/contents/{"/".join(list[7:])}?ref={branch}'
        try:
            # if(username is not None and token is not None):
            headers=None
            if bearer_token is not None:
                headers = {"Authorization": "Bearer {}".format(bearer_token)}
            elif username is not None and bearer_token is not None:
                headers = { 'Authorization' : basic_auth(username, bearer_token)}
            try:    
                response = requests.get(api,headers=headers, timeout=5)
            except requests.exceptions.Timeout:
                logging.info("The request timed out")
                return None    
            # response=requests.get(api)
            json_response=json.loads(response.text)
            content_api=json_response["download_url"]
            try: 
                response_content = requests.get(content_api, timeout=5)
            except requests.exceptions.Timeout:
                logging.info(f"Request to {content_api} timed out")
            except requests.exceptions.RequestException as e:
                logging.info(f"Request failed: {e}")
            
            if(link.split(".")[-1]=="xml"):
                log_dir = str(os.path.join(tempfile.gettempdir(), 'XML'))
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
                file = os.path.join(log_dir, 'XML_' + str(uuid.uuid4()) + '.xml')
                logging.info("GIT XML PATH"+str(file))
            else:
                file=os.path.join(tempfile.gettempdir(),link.split("/")[-1])
                logging.info("GIT FILE PATH"+str(file))
            with open(file, "w+") as f:
                f.write(response_content.text)
            logging.info("FILE IS DOWNLOADED")
            return file
        except Exception as e:                  
            logging.info("Some Error while running the API")

def basic_auth(username, password):
    token = b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
    return f'Basic {token}'

if __name__ == "__main__":
    fetchFileFromGit("GIT:https://github.com/Gemini-Solutions/gempyp/blob/dev/tests/configTest/Gempyp_Test_suite.xml:dev:username:token")


import json
import requests
import logging
import os, re
import sys
import traceback
# import urllib
from gempyp.config import DefaultSettings


# upload_file_api = DefaultSettings.getUrls('bucket-file-upload-api')
# upload_data_api = DefaultSettings.getUrls('bucket-data-upload-api')
# delete_file = DefaultSettings.getUrls('bucket-file-modify')

def upload_to_s3(api=None, tag=None, file=None, data=None, s_run_id=None, folder=None, username=None, bearer_token=None, bridge_token=None):
    logging.info("In Upload section")
    if bearer_token is not None:
        headers = {"Authorization": "Bearer {}".format(bearer_token)}
    elif bridge_token is not None:
        if username is None:
            logging.ERROR("------ Please pass username along with Bridge token -----")
            sys.exit()
        headers = {"bridgeToken": bridge_token, "username": username}
    params = dict()
    if folder is not None:
        params["folder"] = folder
    if tag is not None: 
        params["tag"] = tag
    if s_run_id is not None:
        params["s_run_id"] = s_run_id
    if data is not None:
        logging.info("------ Uploading data to s3 -----")
        try:
            api = DefaultSettings.getUrls('bucket-data-upload-api') if not api else api
            headers["Content-Type"] = "text/plain"
            params["file"] = file
            try:
                data = str(data).encode('utf-8')
            except Exception as e:
                logging.warn(e)
                traceback.print_exc()
            response = requests.post(api, params=params, data=data, headers=headers, timeout=3)
            data = json.loads(response.text)  
            if response.status_code == 200:
                data = json.loads(response.text)    
                data_info = data["data"]
                return data_info
        except Exception as e:
            traceback.print_exc()
            logging.warn(e)
    elif file is not None:
        file = file.split(",")
        files = list()
        logging.info("---- Uploading file to s3 ----")
        api = DefaultSettings.getUrls('bucket-file-upload-api') if not api else api
        for f in file:
            if not os.path.isfile(f) or f == "N.A":
                logging.error("Path of file invalid - " + str(f))
                continue
            files.append(("file", (f.split("\\")[-1],open(f, "rb"),'text/xml')))
        response = requests.post(api, files=files, params=params, headers=headers, timeout=3) 
        data = json.loads(response.text) 
        if response.status_code == 200:
            data = json.loads(response.text)    
            file_info = data["data"]

            return file_info
        


def uploadToS3(api=None, tag=None, file=None, data=None, s_run_id=None, folder=None, username=None, bearer_token=None, bridge_token=None):

    print("*****************we are in upload to s3 function***************")
    
    """need to be make correct just dummy code"""
    logging.info("In Upload section")
    if bearer_token is not None:
        headers = {"Authorization": "Bearer {}".format(bearer_token)}
    elif bridge_token is not None:
        if username is None:
            logging.ERROR("------ Please pass username along with Bridge token -----")
            sys.exit()
        headers = {"bridgeToken": bridge_token, "username": username, "Content-Type": "application/json"}
    body = {}
    if folder is not None:
        body["folder"] = folder
    if s_run_id is not None:
        body["s_run_id"] = s_run_id

    if type(file) == str:
        file = file.split(',')
    elif type(file) == list:
        pass
    else:
        logging.error("file data is inaccurate")
    filename = []
    for each in file:
        filename.append(os.path.basename(each))
    body["files"] = filename 
    body = json.dumps(body)
    response = requests.post(api, data=body, headers=headers, timeout=3)
    try:

        if response.status_code == 200:
            responseData = json.loads(response.text)
            resultData = responseData["data"]
            finalResult = []
            for each in file:
                filename = os.path.basename(each)
                j = None
                for i in range(len(resultData)):
                    j = i
                    files = []
                    if resultData[i].get(filename,None):
                        if not os.path.isfile(each):
                            logging.error("Path of file invalid - " + str(each))
                            continue
                        try:
                            files.append(("file", (os.path.basename(each),open(each, "rb"),'text/csv')))
                        except Exception:
                            print(traceback.print_exc())
                        api = resultData[i].get(os.path.basename(each),None)
                        response = requests.put(api, files=files, headers=headers, timeout=3)
                        if response.status_code == 200: ## this logic is needed to be modified currently support only one file at a time
                            # url = {}
                            # url["url"] = resultData[i].get("Url",None)
                            finalResult.append(resultData[i].get("Url",None))
                            # finalResult.append(url)
                        
                if j :
                    resultData.pop([j])
            return finalResult
        else:
            logging.warn("Some Error Ocurred While Uploading File to S3")
    except Exception:
            print(traceback.print_exc())
        
def download_from_s3(api, bearer_token=None, bridge_token=None, username=None, id=None):
    logging.info("In Download section")
    if bearer_token is not None:
        headers = {"Authorization": "Bearer {}".format(bearer_token)}
    elif bridge_token is not None:
        if username is None:
            logging.ERROR("------ Please pass username along with Bridge token -----")
            sys.exit()
        headers = {"bridgeToken": bridge_token, "username": username}
    params = dict()
    if '?' not in api and id is None:
        logging.error("Please pass download url or the id")
        sys.exit()
    elif '?' in api:
        api_query = api.split('?')
        params = api_query[1][3:]
        api = api_query[0]
        params = {"id": "{}".format(params)}
    elif '?' not in api and id is not None:
        params = {"id": id}
    response = requests.get(api, params=params, headers=headers, timeout=3)  
    return response


def create_s3_link(**kwargs):
    """ creating s3 link to be viewed on file viewer on jewel UI"""
    s3_viewer = DefaultSettings.getUrls('file-viewer')
    if kwargs.get("url", None):
        params = "url=" + kwargs.get("url")
        kwargs.pop("url")
        for key, value in kwargs.items():
            params = f"{params}&{key}={value}"
        return f"{s3_viewer}?{params}"
    return None

# if __name__ == "__main__":
#     print(download_from_s3(api="",username="",bridge_token=""))



import traceback
import requests
import logging
from gempyp.config import DefaultSettings
from gempyp.libs.enums.status import status
import logging
import re
import json
import sys

not_uploaded = []
suite_data = []
flag = False
suite_uploaded = False
list_of_testcase = []
respon = {}

# stat = {k.name: 0 for k in status}
# stat.update({"TOTAL":0})
stat= {**{k.name: 0 for k in status}, **{"TOTAL":0}}
def _getHeaders(bridge_token, user_name):
    """
    for getting the bridgeToken in _sendData method
    """
    return {"Content-Type": "application/json", "bridgeToken": bridge_token, "username": user_name}

def checkingData(run_id, bridge_token, user_name):
    try:
        url = DefaultSettings.getUrls("suite-exe-api") + "?s_run_id=" + run_id
        response = _sendData("", url , bridge_token, user_name, "GET")
        if response.status_code == 200:
            logging.info("Retrying to update and add testcases data present in DB")
            logging.info("Get Request successfull")
            global respon, suite_uploaded, list_of_testcase
            suite_uploaded = True
            respon = response.json()
            # list_of_testcase = [s[:-37] for s in respon['data']['testcase_details']]
            list_of_testcase = [s[:-37] for s in respon['data'].get('testcase_details', [])] if respon['data'].get('testcase_details', None) else []
            return response._content
        else:
            # return "failed"
            return False
    except Exception as e:
        traceback.print_exc()
        # return "failed"
        return False

def sendSuiteData(payload, bridge_token, user_name, mode="POST"):
    """
    for checking the sendSuiteData api response
    """

    try:
        if len(respon) != 0:
            payload = dataAlter(payload)
        payload = noneRemover(payload)
        response = _sendData(payload, DefaultSettings.getUrls("suite-exe-api"), bridge_token, user_name, mode)
        response_message = json.loads(response.text)["message"]
        autoKill = False
        if "New executions not allowed. Either enable AutoKill" in response_message:
            autoKill = True
        if response and response.status_code in [201, 200]:
            global suite_uploaded
            logging.info("Suite data uploaded successfully")
            suite_uploaded = True
            try:
                DefaultSettings.project_id = json.loads(response.text)["data"]["p_id"]
            except Exception as e:
                logging.info(traceback.format_exc())
            if payload in suite_data:
                suite_data.remove(payload)
        elif response and response.status_code == 200:
            logging.info("Suite Data updated Successfully")
        elif response.status_code and autoKill:
            logging.info("New executions not allowed. Either enable AutoKill or abort any previous incomplete executions before new suite runs.")
            sys.exit()
        elif re.search('50[0-9]',str(response.status_code)):
            logging.info("Suite data is not uploaded")
            if payload not in suite_data:
                suite_data.append(payload)
        # else:
        #     logging.info("Some Error From the Client Side, Terminating Execution")
        #     sys.exit()
        else:
            logging.info("Suite data is not uploaded")
            if payload not in suite_data:
                suite_data.append(payload)
        return response.status_code
    except Exception as e:
        logging.error(traceback.format_exc())

def sendTestcaseData(payload, bridge_token, user_name):
    """
    for checking the sendTestCaseData api response
    """
    try:
        method = "POST"
        payload = json.loads(payload)
        payload = attachmentRemover(payload)
        tc_run_id = payload["tc_run_id"]
        if tc_run_id[:-37] in list_of_testcase:
        #checking whether testcase present in previous run 
            payload, method = getTestcase(payload, method, bridge_token, user_name)
        payload = json.dumps(payload)
        response = _sendData(payload, DefaultSettings.getUrls("test-exe-api"), bridge_token, user_name, method)
        ### Applying regex to the response
        x = re.search("already present",response.text,re.IGNORECASE)
        if response.status_code == 201:
            logging.info("data uploaded successfully")
            if payload in not_uploaded:
                not_uploaded.remove(payload)
        elif response.status_code == 200:
            logging.info("data updated successfully")
            if payload in not_uploaded:
                not_uploaded.remove(payload)
    ### code for adding unuploaded testcases
        # elif re.search('50[0-9]',str(response.status_code)):
        #     if payload not in not_uploaded:
        #         not_uploaded.append(payload)
        #         logging.info("Testcase data is not uploaded")
        #         if x != None:
        #             global flag
        #             flag = True
        # else:
        #     logging.info("Some Error From the Client Side, Terminating Execution")
        #     sys.exit()
        elif response.status_code == 412:
            logging.info("Suite is killed on Jewel")
            return response.status_code
        else:
            if payload not in not_uploaded:
                not_uploaded.append(payload)
                logging.info("Testcase data is not uploaded")
                if x != None:
                    global flag
                    flag = True
    except Exception as e:
        logging.error(traceback.format_exc())


def _sendData(payload, url, bridge_token, user_name, method="POST"):
    
    """
    calling the api to upload the data into database
    takes data we need to send(payload),bridgeToken,userName and method as argument
    """
    # Not needed anymore as we will be reuploading the data to db.
    # if DefaultSettings.count > 3:         
    #     logging.warning("Incorrect bridgetoken/username or APIs are down. Skipping Data upload.")
    #     return None
    response = requests.request(
        method=method,
        url=url,
        data=payload,
        headers=_getHeaders(bridge_token, user_name)
    )
    logging.info(f"Response text: {response.text}")
    logging.info(f"status: {response.status_code}")
    logging.info(f"Payload: {payload}")
    return response

def noneRemover(payload):
    data = json.loads(payload)
    for key,value in dict(data).items():
        if value == "-":
            del data[key]
    return json.dumps(data)

# code for changing suite data
def dataAlter(payload):
    payload = json.loads(payload)
    global respon, stat
    # changing start time
    payload['s_start_time'] = respon['data']['s_start_time']
    payload['testcase_info'] = {} if payload['testcase_info'] == "-" else payload['testcase_info']
    if respon['data'].get('testcase_info',None):
        # adding the testcase analytics of both run
        payload['testcase_info'] = {i: payload['testcase_info'].get(i, 0) + respon['data']['testcase_info'].get(i, 0)
        for i in set(payload['testcase_info']).union(respon['data']['testcase_info'])}
        # updating the testcase analytics according to new run
        for key, value in stat.items():
            if key in payload['testcase_info']:
                payload['testcase_info'][key] += value

    sorted_dict = {'TOTAL':0, 'PASS':0, 'FAIL':0}
    sorted_dict.update(payload['testcase_info'])
    payload['testcase_info'] = sorted_dict

    tc_count = 0
    try:
        # tc_count = sum(list(set(payload.get('testcase_info', {}).values()) - set(['TOTAL'])))
        tc_count = sum(payload.get('testcase_info', {}).values()) - payload.get('testcase_info', {}).get("TOTAL", 0)
    except Exception as e:
        traceback.print_exc()

    if tc_count > 0:
        payload['expected_testcases'] = tc_count
    for s in status:
        if sorted_dict.get(s.name, 0) > 0:
            payload['status'] = s.name
            break
    return  json.dumps(payload) 

# code for checking and updating testcase details
def getTestcase(payload, method, bridge_token, user_name):
    global flag

    for i in respon['data']['testcase_details']:

        if respon['data']['testcase_details'] and payload['tc_run_id'][:-37] in i:
            url = DefaultSettings.getUrls("test-exe-api") + "?tc_run_id=" + i
            response = _sendData(" ",url, bridge_token, user_name, "GET")
            if response.status_code == 200:
                logging.info("Testcase get request successfull")
                global flag
                flag = True
                if payload in not_uploaded:
                    not_uploaded.remove(payload)
            res = response.json()
            status = res['data']['status']
            stat[status] -= 1
            payload['tc_run_id'] = i
            stat["TOTAL"] -=1
            method = "PUT"
    return payload, method

def attachmentRemover(payload):
    data = payload.get("steps")
    for i in range(len(data)):
        for key,value in dict(data[i]).items():
            if value == "-":
                del data[i][key]
    payload['steps'] = data
    return payload



import traceback, os, argparse
from gempyp.libs.enums.status import status
from gempyp.plugin.common import common

total_scenario_count = 0

def countingScenarios(filePath):
    with open(filePath, 'r') as file:
        content = file.read()
        lines = content.splitlines()
        scenario_count = 0
        in_scenario_block = False
        for line in lines:
            line = line.strip()
            if line.startswith("Scenario"):
                in_scenario_block = True
                scenario_count += 1
            elif in_scenario_block and line.startswith("#"):
                # If we're in a scenario block and the line starts with '#', it's a commented scenario.
                scenario_count -= 1
            elif line == "":
                # Reset the flag when encountering an empty line to exit the scenario block.
                in_scenario_block = False
    return scenario_count


def before_all(context):
    # scenario_count = len(feature.scenarios)
    # suite_dict = {"expected-testcases":0}
    feature_directory = 'features'  # Change this to the directory where your feature files are located
    try:
        filePath = None
        parser = argparse.ArgumentParser()
        parser.add_argument('-filePath','-k',dest='filePath',type=str, required=False)
        args = parser.parse_args()
        config = vars(args)
        filePath = config.get('filePath',None)
    except Exception:
        print(traceback.print_exc())



    # Iterate through all feature files
    total_scenarios = 0
    if filePath:
        total_scenarios = countingScenarios(filePath)
    else:
        for root, dirs, files in os.walk(os.path.abspath(feature_directory)):
            for filename in files:
                if filename.endswith('.feature'):
                    filePath = os.path.join(root, filename)
                    total_scenarios += countingScenarios(filePath)
                    # Parse the feature file to count scenarios
                
    suite_dict = {"expected-testcases":total_scenarios}
    common.makeSuiteDetails(suite_dict)
    common.sendSuiteData()
    # before_feature(context, feature)

def after_all(context):
    common.createReport()
    

def before_scenario(context, scenario):
    test = {"testcaseName":scenario.name}
    common.getTestcaseData(test)


def after_scenario(context, scenario):
    feature_filename = context._runner.feature.filename
    common.reporter.addMisc("feature",feature_filename)
    common.send_testcase_data()


def after_step(context, step):

    txt_tb = u"".join(traceback.format_tb(step.exc_traceback))
    try:
        if str(step.status) == "Status.failed":
            common.reporter.addRow(step.name,str(step.exception).split('}')[0],common.reporter.Status.FAIL)
        elif str(step.status) == "Status.passed":
            common.reporter.addRow(step.name,step.keyword.strip(),common.reporter.Status.PASS)
        else:
            txt_tb = u"".join(traceback.format_tb(step.exc_traceback))
            common.reporter.addRow(step.name,str(step.exception).split('}')[0],common.reporter.Status.ERR)
    except Exception:
        common.reporter.addRow(step.name,traceback.print_exc(),status.ERR)





import uuid
from gempyp.libs.gem_s3_common import uploadToS3, create_s3_link
from gempyp.config import DefaultSettings
import mysql.connector
import os
import uuid
import configparser
from typing import Dict
import pandas as pd
from gempyp.engine.baseTemplate import TestcaseReporter as Base
from gempyp.libs.common import moduleImports
from gempyp.libs.enums.status import status
from gempyp.libs.common import readPath
from gempyp.dv.dvReporting import writeToReport
import traceback
import pandas as pd
import logging
import math
import numpy
import pg8000
from gempyp.dv.dvObj import DvObj
from gempyp.libs.common import download_common_file,get_reason_of_failure
from gempyp.engine.runner import getError
from gempyp.libs import common
from gempyp.dv.dvDataframe import Dataframe
from gempyp.dv.dvCompare import df_compare
from gempyp.dv.dfOperations import dateFormatHandling, columnCompare, skipColumn
import re
import json
import ast


class DvRunner(Base):

    def __init__(self, data):

        self.data = data
        self.configData: Dict = self.data.get("config_data")
        self.logger = data["config_data"]["LOGGER"] if "LOGGER" in data["config_data"].keys(
        ) else logging
        self.logger.info(
            "---------------------Inside DV FRAMEWORK------------------------")
        self.logger.info(
            f"-------Executing testcase - \"{self.data['config_data']['NAME']}\"---------")
        if data.get("default_urls", None):
            # only for optimized mode, urls not shared between processes
            DefaultSettings.urls.update(data.get("default_urls"))
        self.setVars()
        self.logger.info(
            "--------------------Report object created ------------------------")
        self.reporter = Base(project_name=self.project,
                            testcase_name=self.tcname)

    def dvEngine(self):

        self.sourceCred = None
        self.targetCred = None
        try:
            try:
                if "SOURCE_DB" or "TARGET_DB" in self.configData:
                    if "DB_CONFIG_PATH" in self.configData:
                        configPath = self.configData["DB_CONFIG_PATH"]
                        self.configur = configparser.RawConfigParser()
                        config = readPath(configPath)
                        self.configur.read(config)
                    else:
                        self.configur = None
            except Exception as e:
                self.reporter.addRow(
                    "Config file", "Path is not correct", status.ERR)
                raise Exception(e)
            if 'SOURCE_DB' or 'TARGET_DB' in self.configData:
                self.dbConnParser()

            if "KEYS" in self.configData:
                self.keys = self.configData["KEYS"].split(',')
                self.reporter.addMisc("KEYS", ", ".join(self.keys))
            else:
                self.reporter.addRow(
                    "Checking keys", "User given keys tag not found", status.ERR)
                raise Exception("User given keys not found")

            obj = Dataframe()
            self.source_df, self.source_columns = Dataframe.getSourceDataFrame(
                obj, self.data, self.logger, self.reporter, self.sourceCred)
            self.target_df, self.target_columns = Dataframe.getTargetDataframe(
                obj, self.data, self.logger, self.reporter, self.targetCred)
            
            if "BEFORE_FILE" in self.configData:
                self.beforeMethod()
            # for columns mapping
            if "COLUMN_MAP" in self.configData:
                self.target_df.rename(columns=ast.literal_eval(
                    self.configData["COLUMN_MAP"]), inplace=True)
                self.target_columns = list(self.target_df.columns)
            self.matchKeys(self.source_columns, "SOURCE")
            self.matchKeys(self.target_columns, "TARGET")

            if 'SKIP_COLUMN' in self.configData:
                self.source_df, self.target_df = skipColumn(self.configData.get('SKIP_COLUMN'),self.source_df,self.target_df,self.keys,self.reporter)
                self.source_columns = list(self.source_df.columns)
                self.target_columns = list(self.target_df.columns)
            #calling getDuplicatekeysDf function to get dataframe of duplicate keys 
            source_duplicates_df, src_dup_len = self.getDuplicateKeysDf(self.source_df, "SOURCE")
            target_duplicates_df, tgt_dup_len = self.getDuplicateKeysDf(self.target_df, "TARGET")
            dup_keys_length = src_dup_len + tgt_dup_len
            duplicate_keys_df = pd.concat([source_duplicates_df , target_duplicates_df ], axis=0, ignore_index=True)

            if self.source_columns == self.target_columns:
                pass
            else:
                self.logger.info(
                    "--------Same Column not Present in Both Table--------")
                self.reporter.addRow(
                    "Same columns in table", "Not found", status.ERR)
                raise Exception("Same Columns Not Found")

            """deleting duplicates from df and keeping last ones"""
            self.logger.info("Removing Duplicates Rows")
            self.source_df.drop_duplicates(
                subset=self.keys, keep='last', inplace=True)
            self.target_df.drop_duplicates(
                subset=self.keys, keep='last', inplace=True)
            
            #calling compare column
            if "COMPARE_COLUMN" in self.configData:
                compare_column = self.configData.get("COMPARE_COLUMN",'').split(',')
                self.source_df, self.target_df = columnCompare(self.source_df, self.target_df, self.keys, compare_column)
            # hadling case insensitivity
            if 'MATCH_CASE' in self.configData:
                self.matchCase()

            # date format handling
            self.source_df, self.target_df = dateFormatHandling(self.source_df, self.target_df)
            
            #checking column compare
            value_dict, key_dict, keys_length = df_compare(
                self.source_df, self.target_df, self.keys, self.logger, self.reporter, self.configData)
            
            self.writeExcel(value_dict, key_dict, keys_length, duplicate_keys_df, dup_keys_length)
            self.reporter.finalizeReport()
            output = writeToReport(self)
            return output, None
        except Exception as e:
            self.reporter.addMisc(
                    "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e))
            self.reporter.addRow("Stopping execution", get_reason_of_failure(traceback.format_exc(), e), status.ERR)
            self.logger.error(str(e))
            traceback.print_exc()
            output = writeToReport(self)
            return output, None

    def dbConnParser(self):
        """checking db type and fetching credentials or connection details accordingly"""
        try:
            if 'SOURCE_DB' in self.configData:
                if 'SOURCE_CONN' in self.configData: 
                    is_dict = True
                    try:
                        db_details = ast.literal_eval(self.configData["SOURCE_DB"])
                    except Exception:
                        is_dict = False
                    if is_dict == True:
                        db_type = db_details.get('type')
                    else:
                        db_type = self.configData["SOURCE_DB"]
                    if db_type.lower() == 'custom':
                        self.sourceCred = self.configData['SOURCE_CONN']
                    else:
                        if re.search("^{.*}$", self.configData["SOURCE_CONN"]):
                            self.sourceCred = ast.literal_eval(self.configData["SOURCE_CONN"])
                        else:
                            self.logger.info("----Parsing the Config File For Source Connections----")
                            self.logger.info("--- get config file from config file path ---")
                            self.sourceCred = dict(self.configur.items(
                                self.configData["SOURCE_CONN"]))
                            self.logger.info("----getting source cred----")
                            self.sourceCred=self.parseConfigDict(self.sourceCred)
            
            if 'TARGET_DB' in self.configData:
                if 'TARGET_CONN' in self.configData:
                    is_dict = True
                    try:
                        db_details = ast.literal_eval(self.configData["TARGET_DB"])
                        self.logger.info("----Parsing the Config File For Target Connections----")
                    except Exception:
                        is_dict = False
                    if is_dict == True:
                        db_type = db_details.get('type')
                    else:
                        db_type = self.configData["TARGET_DB"]
                    
                    if db_type.lower() == 'custom':
                        self.targetCred = self.configData['TARGET_CONN']
                    else:
                        if re.search("^{.*}$", self.configData["TARGET_CONN"]):
                            self.targetCred = ast.literal_eval(self.configData["TARGET_CONN"])
                        else:
                            self.targetCred = dict(self.configur.items(
                                self.configData["TARGET_CONN"]))
                            self.targetCred=self.parseConfigDict(self.targetCred)

            if self.configur != None:
                self.logger.warn("Config file is parsed successfully")
                self.reporter.addRow(
                    "Parsing db conf", "Parsing of db config is successful", status.PASS)
        except Exception as e:
            self.logger.error("Inside exception of dbConnParser")
            self.reporter.addRow(
                "Parsing db conf", "Exception occurred", status.ERR)
            self.reporter.addMisc(
                "REASON OF FAILURE", common.get_reason_of_failure(traceback.format_exc(), e))
            raise Exception(e)

    def matchKeys(self, columns, db):

        try:
            self.logger.info(f"Matching Keys in {db} DB")
            key = []
            for i in self.keys:
                if i not in columns:
                    key.append(i)
            if len(key) == 0:
                self.reporter.addRow(
                    f"Matching given keys in {db}", f"Keys are present in {db}", status.PASS)
                self.logger.info(f"Given Keys are Present in {db} DB")
            else:
                raise (f"Keys are not Present in {db}")
        except Exception as e:
            keyString1 = ", ".join(key)
            self.reporter.addRow(
                f"Matching given keys in {db}", "Keys: " + keyString1 + f" are not present in {db}db", status.ERR)
            self.logger.info("------Given Keys are not present in DB------")
            raise Exception("Keys not Present in DB")

    def setVars(self):
        """
        For setting variables like testcase name, output folder etc.
        """
        self.data["config_data"]=self.parseConfig(self.data["config_data"])
        self.default_report_path = os.path.join(os.getcwd(), "pyprest_reports")

        self.data["REPORT_LOCATION"] = self.data.get("REPORT_LOCATION", self.default_report_path)
        if self.data["REPORT_LOCATION"].strip(" ") == "":
            self.data["REPORT_LOCATION"] = self.default_report_path

        self.project = self.data["PROJECT_NAME"]
        self.tcname = self.data["config_data"]["NAME"]
        self.env = self.data["ENVIRONMENT"]
        self.category = self.data["config_data"].get("CATEGORY", None)

    def writeExcel(self, valDict, keyDict, keys_length, duplicate_keys_df, dup_keys_len):

        try:
            logging.info("----------in write excel---------")
            key_check = len(keyDict["Reason-of-Failure"])
            value_check = len(valDict["Reason-of-Failure"])
            self.logger.info("In write Excel Function")
            outputFolder = self.data['REPORT_LOCATION']

            unique_id = uuid.uuid4()
            excelPath = os.path.join(
                outputFolder, self.configData['NAME']+str(unique_id)+'.csv')
            excel = os.path.join(
                ".", "testcases", self.configData['NAME']+str(unique_id)+'.csv')
            self.value_df = pd.DataFrame(valDict)
            self.keys_df = pd.DataFrame(keyDict)
            if "AFTER_FILE" in self.configData:
                self.afterMethod()
            if (key_check + value_check + dup_keys_len) == 0:
                self.reporter.addRow("Data validation report",
                                    "No misMatch value found", status=status.PASS)
                self.logger.info("----No MisMatch Value Found----")
            else:
                self.logger.info("----Adding Data to Excel----")
                # with pd.ExcelWriter(excelPath) as writer1:
                #     if key_check == 0:
                #         pass
                #     else:
                #         self.keys_df.to_excel(
                #             writer1, sheet_name='key_difference', index=False)
                #     if value_check == 0:
                #         pass
                #     else:
                #         self.value_df.to_excel(
                #             writer1, sheet_name='value_difference', index=False)
                df = pd.concat([self.value_df, self.keys_df, duplicate_keys_df ], axis=0, ignore_index=True)
                df_columns = set(df.columns)
                fixed_columns = {"Column-Name","Source-Value","Target-Value","Reason-of-Failure"}
                diff_columns = df_columns - fixed_columns
                new_order = list(diff_columns)+["Column-Name","Source-Value","Target-Value","Reason-of-Failure"]
                df = df[new_order]
                df.to_csv(excelPath, index=False,header=True)
                print(excelPath)
                s3_url = None
                try:
                    s3_url = uploadToS3(DefaultSettings.urls["data"].get("pre-signed",None), bridge_token=self.data["SUITE_VARS"]
                                            ["bridge_token"], tag="protected", username=self.data["SUITE_VARS"]["username"], file=excelPath,s_run_id=self.data.get("S_RUN_ID"))[0]
                except Exception as e:
                    print(traceback.print_exc())
                    logging.warn(e)
                if not s3_url:
                    self.reporter.addRow("Data validation report", f"Matched keys: {keys_length['common_keys']}, Keys only in source: {keys_length['keys_only_in_src']}, Keys only in target: {keys_length['keys_only_in_tgt']}, Mismatched cells: {value_check}, Duplicate keys: {dup_keys_len}, DV result file:", status=status.FAIL,Attachment=[s3_url])
                else:
                    self.reporter.addRow("Data Validation Report", f"Matched keys: {keys_length['common_keys']}, Keys only in source: {keys_length['keys_only_in_src']}, Keys only in target: {keys_length['keys_only_in_tgt']}, Mismatched cells: {value_check}, Duplicate keys: {dup_keys_len}, DV result file:", status=status.FAIL,Attachment=[s3_url])

                self.reporter.addMisc("REASON OF FAILURE", str(
                    f"Mismatched Keys: {key_check},Mismatched Cells: {value_check}"))
            self.reporter.addMisc("common Keys", str(keys_length['common_keys']))
            self.reporter.addMisc("Keys Only in Source",
                                str(keys_length['keys_only_in_src']))
            self.reporter.addMisc("Keys Only In Target",
                                str(keys_length['keys_only_in_tgt']))
            self.reporter.addMisc("Mismatched Cells", str(value_check))
        except Exception as e:
            self.reporter.addMisc(
                    "REASON OF FAILURE", get_reason_of_failure(traceback.format_exc(), e))
            self.reporter.addRow("Writing data into csv","Exception occured",status.ERR)

    def beforeMethod(self):
        """This function
        -checks for the before file tag
        -stores package, module,class and method
        -runs before method if found
        -takes all the data from before method and updates the self object"""

        # check for before_file
        self.logger.info("CHECKING FOR BEFORE FILE___________________________")

        file_str = self.data["config_data"].get("BEFORE_FILE", "")
        if not file_str or file_str == "" or file_str == " ":
            self.logger.info(
                "BEFORE FILE NOT FOUND___________________________")
            self.reporter.addRow(
                "Searching for before file steps", "No before file steps found", status.INFO)

            return
        self.reporter.addRow("Searching for before file steps",
                            "Searching for before file", status.INFO)

        file_name = file_str.split("path=")[1].split(",")[0]
        if "CLASS=" in file_str.upper():
            class_name = file_str.split("class=")[1].split(",")[0]
        else:
            class_name = ""
        if "METHOD=" in file_str.upper():
            method_name = file_str.split("method=")[1].split(",")[0]
        else:
            method_name = "before"

        self.logger.info("Before file path:- " + file_name)
        self.logger.info("Before file class:- " + class_name)
        self.logger.info("Before file mthod:- " + method_name)
        try:
            file_path = download_common_file(
                file_name, self.data.get("SUITE_VARS", None))
            file_obj = moduleImports(file_path)
            self.logger.info("Running before method")
            obj_ = file_obj
            before_obj = DvObj(
                project=self.project,
                source_df=self.source_df,
                target_df=self.target_df,
                source_columns=self.source_columns,
                target_columns=self.target_columns,
                keys=self.keys,
                env=self.env,
                reporter=self.reporter 
            )
            if class_name != "":
                obj_ = getattr(file_obj, class_name)()
            fin_obj = getattr(obj_, method_name)(before_obj)
            self.extractBeforeObj(fin_obj)

        except Exception as e:
            self.logger.info(traceback.format_exc())
            self.reporter.addRow(
                "Executing before method", f"Some error occurred while searching for before method- {str(e)}", status.ERR)

    def afterMethod(self):
        """This function
        -checks for the after file tag
        -stores package, module,class and method
        -runs after method if found
        -takes all the data from after method and updates the self object"""

        self.logger.info("CHECKING FOR AFTER FILE___________________________")

        file_str = self.data["config_data"].get("AFTER_FILE", "")
        if not file_str or file_str == "" or file_str == " ":
            self.logger.info("AFTER FILE NOT FOUND___________________________")
            self.reporter.addRow(
                "Searching for after file steps", "No file path found", status.INFO)
            return

        self.reporter.addRow("Searching for after file",
                            "Searching for file", status.INFO)

        file_name = file_str.split("path=")[1].split(",")[0]
        if "CLASS=" in file_str.upper():
            class_name = file_str.split("class=")[1].split(",")[0]
        else:
            class_name = ""
        if "METHOD=" in file_str.upper():
            method_name = file_str.split("method=")[1].split(",")[0]
        else:
            method_name = "after"
        self.logger.info("After file path:- " + file_name)
        self.logger.info("After file class:- " + class_name)
        self.logger.info("After file mthod:- " + method_name)
        try:
            file_path = download_common_file(
                file_name, self.data.get("SUITE_VARS", None))
            file_obj = moduleImports(file_path)
            self.logger.info("Running After method")
            obj_ = file_obj
            after_obj = DvObj(
                project=self.project,
                value_df=self.value_df,
                keys_df=self.keys_df,
                env=self.env,
                reporter=self.reporter 
            )
            if class_name != "":
                obj_ = getattr(file_obj, class_name)()
            fin_obj = getattr(obj_, method_name)(after_obj)
            self.extractAfterObj(fin_obj)
        except Exception as e:
            self.reporter.addRow(
                "Executing after method", f"Some error occurred while searching for after method- {str(e)}", status.ERR)

    def extractBeforeObj(self, obj):
        """To ofload the data from pyprest obj helper, assign the values back to self object"""

        self.reporter = obj.object
        self.project = obj.project
        self.source_columns = obj.source_columns
        self.target_columns = obj.target_columns
        self.source_df = obj.source_df
        self.target_df = obj.target_df
        self.keys = obj.keys
        self.env = obj.env
        self.reporter = obj.reporter

    def extractAfterObj(self, obj):

        self.reporter = obj.object
        self.project = obj.project
        self.value_df = obj.value_df
        self.keys_df = obj.keys_df
        self.env = obj.env
        self.reporter = obj.reporter

    def matchCase(self):
        try:
            columns = self.configData['MATCH_CASE'].split(',')
            for i in columns:
                self.source_df[i] = self.source_df[i].apply(str.lower)
                self.target_df[i] = self.target_df[i].apply(str.lower)
        except Exception as e:
            self.reporter.addRow("Trying to match case",common.get_reason_of_failure(traceback.format_exc(), e), status.ERR)
            raise e
        
    def getDuplicateKeysDf(self, df, type):

        self.logger.info("Checking Dulicates Keys")
        dup_df = df[df[self.keys].duplicated(keep=False)]

        dup_keys_df = dup_df[self.keys]
        dup_length = len(dup_keys_df)
        dup_keys_df.drop_duplicates(
                keep='last', inplace=True)
        dup_keys_df['Reason-of-Failure'] = f'Duplicate Key in {type}'
        if len(dup_keys_df['Reason-of-Failure']) > 0:
            self.reporter.addRow(f"Checking for duplicates keys in {type}",f"Found duplicate keys in {type}",status.FAIL)
        return dup_keys_df, dup_length
    
    def parseConfig(self,config):
        pattern = r"ENV.([a-zA-Z0-9_]+)"
        for key in config.keys():
            value=config.get(key)
            if(type(value)!=logging.Logger):
                if re.search("mysql.connector.connect",value) or re.search("^{",value):
                    config[key] = re.sub(pattern, lambda match: f"'{os.environ.get(match.group(1), '')}'", value)
                else:
                    if("ENV." in value):
                        config[key]=os.environ.get(value.replace("ENV.",""))
        return config
    
    def parseConfigDict(self,conf):
        for key in conf.keys():
            if("ENV." in conf.get(key) and os.environ.get(conf.get(key).replace("ENV.",""))):
                conf[key]=os.environ.get(conf.get(key).replace("ENV.",""))
        return conf
                



import json
import ast
from gempyp.engine import dataUpload
import os
import logging
from cryptography.fernet import Fernet
from gempyp.config.DefaultSettings import encrypt_key, default_baseurl, checkUrl
import sys
import re
import traceback
from gempyp.libs.gem_s3_common import upload_to_s3, create_s3_link
def dataUploader(file_path, bridge_token):

    try:
        # Reading the file sent by the user
        f = open(file_path)
        encData = f.read()
        f.close()
        #decryping the file here
        fernet = Fernet(encrypt_key)
        data = fernet.decrypt(encData).decode()
        data = ast.literal_eval(data)
        if bridge_token:
            data['bridge_token'] = bridge_token
        if not data['base_url']:
            data['base_url'] = checkUrl(default_baseurl)
        print("Base Url Used:",data['base_url'])
        response = dataUpload._sendData(" ",data['base_url'], data['bridge_token'], data['user_name'],"GET")
        if response.status_code == 200:
            url_enter_point = response.json()
            data["urls"] = url_enter_point["data"]  
            try:  
                if 'suite_data' in data:
                    data['suite_data'][-1] = ast.literal_eval(data['suite_data'][-1])
                    for i in range(len(data['suite_data'][-1]['meta_data'])):
                        if 'CONFIG_S3_URL' in data['suite_data'][-1]['meta_data'][i].keys():
                            xml_path = data['suite_data'][-1]['meta_data'][i]['CONFIG_S3_URL']
                            index = i
                            break
                    #trying to upload suite data 
                    if not re.search("^https://.", xml_path):
                        try:
                            xml_s3_url = upload_to_s3(data["urls"]["bucket-file-upload-api"], bridge_token=data['bridge_token'], username=data['user_name'], file=xml_path)[0]["Url"]
                            print("XML S3 URL", xml_s3_url)
                            data['suite_data'][-1]['meta_data'][i]['CONFIG_S3_URL'] = xml_s3_url
                        except Exception:
                            print('Some Problem Occured while uploading xml file on S3')
                    data['suite_data'][-1] = json.dumps(data['suite_data'][-1])
                    payload = str(data['suite_data'][-1])
                    method = "POST"
                    if len('suite_data') == 1:
                        method = "PUT" 
                    response = dataUpload._sendData(payload, data['urls']['suite-exe-api'],data['bridge_token'], data['user_name'], method)
                    print("Suite data API Response")
                    print(response.status_code)
                    if response.status_code == 201 or response.status_code == 200:
                        logging.info('Suite Data is Uploaded Successfully')
                        d = ast.literal_eval(data['suite_data'][0])
                        s_run_id = d['s_run_id']
                        print('Jewel Link',f"{data['urls']['jewel-url']}/#/autolytics/execution-report?s_run_id={s_run_id}")
                        del data['suite_data']
                    else:
                        print('Suite data is not Uploaded, therefore terminating the further execution')
                        sys.exit()
            except Exception:
                traceback.print_exc()
                sys.exit()
            try:       
                if 'testcases' in data:
                    # trying to upload testcase
                    i = -1
                    loop_count = 1
                    testcases_length = len(data['testcases'])
                    while loop_count <= testcases_length:
                        loop_count += 1
                        data['testcases'][i] = json.loads(data['testcases'][i])
                        if not re.search("^https://.", data['testcases'][i]['log_file']):
                            try:
                                print('Uploading Testcases Log File to S3')
                                s3_log_file_url = create_s3_link(url=upload_to_s3(data["urls"]["bucket-file-upload-api"], bridge_token=data['bridge_token'], username=data['user_name'], file=data['testcases'][i]['log_file'])[0]["Url"])
                                data['testcases'][i]['user_defined_data']['LOG_FILE'] = f'<a href="{s3_log_file_url}" target=_blank>view</a>'
                            except Exception:
                                print('some problem occured while uploading testcase log file')
                        if data['testcases'][i]['product_type'] == 'GEMPYP-DV':
                            if not re.search("^https://.", data['testcases'][i]['steps'][-1]["Attachment"][-1]):
                               print('Uploading DV Result File to S3')
                               s3_result_file_url = create_s3_link(url=upload_to_s3(data["urls"]["bucket-file-upload-api"], bridge_token=data['bridge_token'], username=data['user_name'], file=data['testcases'][i]['steps'][-1]["Attachment"][-1])[0]["Url"]) 
                               data['testcases'][i]['steps'][-1]["Attachment"][-1] = s3_result_file_url
                        data['testcases'][i] = json.dumps(data['testcases'][i])    
                        response = dataUpload._sendData(data['testcases'][i], data['urls']['test-exe-api'],data['bridge_token'], data['user_name'])
                        if response.status_code == 200 or response.status_code == 201:
                            print("Testcase is uploaded successfully")
                            del data['testcases'][i]
            except Exception:
                traceback.print_exc()
                                
            if 'suite_data' not in data and len(data['testcases']) == 0:
                #removing the file in case of all the testcase and suite data uploaded successfully
                print("Removing File")
                os.remove(file_path)
            else:
                # Rewriting the data in json file at same location
                print("Updating File")
                data = str(data)
                fernet = Fernet(encrypt_key)
                data = fernet.encrypt(data.encode())
                with open(file_path,'wb') as w:
                    w.write(data)
                # data = json.loads(data)
        elif re.search('50[0-9]',str(response.status_code)):
            print("There is some problem on server side, Please try again after sometime")
        else:
            print("Some Error From the Client Side maybe Username or Bridgetoken, Therefore Terminating Execution")
    except Exception:
        traceback.print_exc()


# from behave import before_scenario, 
# Feature-level Hooks
# @before.feature
import traceback
from gempyp.libs.enums.status import status
from gempyp.plugin.common import common
import time
# from gempyp.plugin.common import common.reporter

def before_feature(context, feature):
    scenario_count = len(feature.scenarios)
    suite_dict = {"expected-testcases":scenario_count}
    common.makeSuiteDetails(suite_dict)
    common.sendSuiteData()

def after_feature(context, feature):
    common.createReport()

def before_scenario(context, scenario):
    test = {"testcaseName":scenario.name}
    common.getTestcaseData(test)

# @after.scenario
def after_scenario(context, scenario):
    common.send_testcase_data()

# Step-level Hooks
# @before.step
def before_step(context, step):
    pass

# @after.step
def after_step(context, step):
    try:
        if str(step.status) == "Status.failed":
            common.reporter.addRow(step.name,step.keyword.strip(),common.reporter.Status.FAIL)
        elif str(step.status) == "Status.passed":
            common.reporter.addRow(step.name,step.keyword.strip(),common.reporter.Status.PASS)
        else:
            common.reporter.addRow(step.name,step.keyword.strip(),common.reporter.Status.ERR)
    except Exception:
        common.reporter.addRow(step.name,traceback.print_exc(),status.ERR)




from calendar import different_locale
from ctypes import util
import logging as logger
import json
import traceback
from numpy import isin
from gempyp.libs.enums.status import status
from gempyp.pyprest.utils import getKeyList 




class legacyApiComparison:
    def __init__(self,pyprest_obj):
        self.pyprest_obj = pyprest_obj
        self.legacy_api_response = self.pyprest_obj.__dict__["legacy_res"]
        self.current_api_response = self.pyprest_obj.__dict__["res_obj"]
        self.current_status_code = self.current_api_response.status_code
        self.legacy_status_code = self.legacy_api_response.status_code
        self.legacy_response_body = json.loads(self.legacy_api_response.response_body)
        self.current_response_body = json.loads(self.current_api_response.response_body)
        self.legacy_response_keys = getKeyList().getKeys(self.legacy_response_body)
        self.current_response_keys = getKeyList().getKeys(self.current_response_body)
        self.logger = self.pyprest_obj.logger
        self.reporter = self.pyprest_obj.reporter

    
    def compare_all(self):
        try:
            self.pyprest_obj.reporter.addRow("Executing Post Assertion check", "Running Assertion on legacy and current API Responses", status.INFO, CURRENT_API="Results | For Current API",LEGACY_API="Results | For Legacy API")
            if compareStatusCodes(self.legacy_status_code, self.current_status_code):
                title = "Comparing status codes of both the responses."
                description = "Both are Equal"
                _status = status.PASS
                legacy_desc = f"LEGACY RESPONSE CODE: {self.legacy_status_code}"
                current_desc = f"CURRENT RESPONSE CODE: {self.current_status_code}"
                self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)

                if compareTypeOfResponses(self.legacy_response_body, self.current_response_body):
                    title = "Comparing type of both the responses."
                    description = "Both responses are of same type"
                    _status = status.PASS
                    legacy_desc = f"LEGACY RESPONSE TYPE: {type(self.legacy_response_body).__name__}"
                    current_desc = f"CURRENT RESPONSE TYPE: {type(self.current_response_body).__name__}"
                    self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                   
                    if compareNumberOfKeys(self.legacy_response_keys, self.current_response_keys):
                        title = "Comparing Number of keys in responses"
                        description = "Both have equal number of keys"
                        _status = status.PASS
                        legacy_desc = f"NUMBER OF KEYS IN LEGACY RESPONSE: {len(self.legacy_response_keys)}"
                        current_desc = f"NUMBER OF KEYS IN CURRENT RESPONSE: {len(self.current_response_keys)}"
                        self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                        
                        if compareKeys(self.legacy_response_keys, self.current_response_keys):
                            title = "Comparing the set of keys in responses"
                            description = "Both responses have same set of keys"
                            _status = status.PASS
                            legacy_desc = f"LIST OF KEYS IN LEGACY RESPONSE:  {self.legacy_response_keys}"
                            current_desc = f"LIST OF KEYS IN CURRENT RESPONSE: {self.current_response_keys}"
                            self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                            
                            if isinstance(self.current_response_body,list) and isinstance(self.legacy_response_body,list):
                                
                                if len(self.current_response_body) == len(self.legacy_response_body):
                                    self.reporter.addRow("Comparing | Record Count of both responses", "Both responses have same count of records", status.PASS, CURRENT_API="Record Count: {}".format(len(self.current_response_body)), LEGACY_API="Record Count: {}".format(len(self.legacy_response_body)))
                                    
                                    for i in range(len(self.legacy_response_body)):
                                        compareResponses(self.reporter, self.current_response_body[i], self.legacy_response_body[i])                                        
                                
                                else:
                                    self.reporter.addRow("Comparing | Record Count of both responses", "Record Count MissMatch", status.FAIL, CURRENT_API="Record Count: {}".format(len(self.current_response_body)), LEGACY_API="Record Count: {}".format(len(self.legacy_response_body)))
                                    self.reporter.addMisc("REASON OF FAILURE", "Record Count Missmatch")
                            
                            elif isinstance(self.current_response_body,dict) and isinstance(self.legacy_response_body,dict):
                                compareResponses(self.reporter, self.current_response_body, self.legacy_response_body)
                       
                        else:
                            title = "Comparing the number of keys in responses"
                            description = "Both responses do not have same set of keys"
                            _status = status.FAIL
                            legacy_desc = f"LIST OF KEYS IN LEGACY RESPONSE: {self.legacy_response_keys}"
                            current_desc = f"LIST OF KEYS IN CURRENT RESPONSE: {self.current_response_keys}"
                            self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                            self.reporter.addMisc("REASON OF FAILURE", "Set of keys do not match")
                    
                    else:
                        title = "Comparing Count of keys in responses"
                        description = "Both do not have equal number of keys."
                        _status = status.FAIL
                        legacy_desc = f"NUMBER OF KEYS IN LEGACY RESPONSE: {len(self.legacy_response_keys)}"
                        current_desc = f"NUMBER OF KEYS IN CURRENT RESPONSE: {len(self.current_response_keys)}"
                        self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                        self.reporter.addMisc("REASON OF FAILURE", "Key Count MissMatch")

                else:
                    title = "Comparing type of both the responses."
                    description = "Both responses are not of same type"
                    _status = status.FAIL
                    legacy_desc = f"LEGACY RESPONSE TYPE: {type(self.legacy_response_body).__name__}"
                    current_desc = f"CURRENT RESPONSE TYPE: {type(self.current_response_body).__name__}"
                    self.logger.warning("---------- Missmatch in responses status codes. ")
                    self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                    self.reporter.addMisc("REASON OF FAILURE", "Response Type Missmatch")

            else:
                title = "Comparing status codes of both the responses."
                description = "Both are not Equal"
                _status = status.FAIL
                legacy_desc = f"LEGACY RESPONSE CODE:  {self.legacy_status_code}"
                current_desc = f"CURRENT RESPONSE CODE:  {self.current_status_code}"
                self.reporter.addRow(title, description, _status, CURRENT_API=current_desc, LEGACY_API=legacy_desc)
                self.reporter.addMisc("REASON OF FAILURE", "Response status code missmatch")
                self.logger.warning("---------- Missmatch in responses status codes. ")

        except Exception as e:
            self.logger.info(traceback.format_exc())
            self.reporter.addMisc("REASON OF FAILURE", f"Some error occurred while sending request- {str(e)}")
            raise Exception("Error occured while sending request - {0}".format(str(e)))



def compareResponses(reporter, current_response, legacy_response):
    if len(current_response)>0 and len(legacy_response)>0:
        
        for (legacy_key, legacy_value), (current_key, current_value) in zip(legacy_response.items(), current_response.items()):
        
            if isinstance(legacy_value, str) and isinstance(current_value, str): 
                responseReporter(reporter, legacy_key, current_key, legacy_value, current_value)
        
            elif isinstance(legacy_value,int) or isinstance(current_value,int):
                responseReporter(reporter, legacy_key, current_key, legacy_value, current_value)                                     
        
            elif isinstance(legacy_value,float) or isinstance(current_value,float):
                responseReporter(reporter, legacy_key, current_key, legacy_value, current_value) 
        
            elif isinstance(legacy_value,bool) or isinstance(current_value,bool):
                responseReporter(reporter, legacy_key, current_key, legacy_value, current_value)
            
            # elif isinstance(legacy_value,None) or isinstance(current_value,None):  #yet to handle 

            elif isinstance(legacy_value, list) or isinstance(current_response,list):
                listComparator(reporter,legacy_key, current_key, legacy_value, current_value)
        
            elif isinstance(legacy_value,dict) or isinstance(current_value,dict):
                compareResponses(reporter, current_value, legacy_value)


def listComparator(reporter,legacy_key, current_key, legacy_value, current_value):
    title = ""
    description = ""
    _status = ""
    current_info = ""
    legacy_info = ""
    if legacy_key == current_key:
        title = f"Comparing | {legacy_key} for Legacy & Current API"
        current_info = f"value = {legacy_value}"
        legacy_info = f"Value = {current_value}"
        if len(legacy_value) != len(current_value) and len(legacy_value)>0 and len(current_value)>0:
            title = "Comparing | record count of {} for legacy and current API".format(legacy_key)
            description = "Record count is not equal"
            _status = status.FAIL
            reporter.addRow(title,description,_status,CURRENT_API=len(current_value), LEGACY_API=len(legacy_value))
            reporter.addMisc("REASON OF FAILURE", "Record count not equal")
        elif len(legacy_value) == len(current_value) and len(legacy_value)>0 and len(current_value)>0:
            title = "Comparing | record count of {} for legacy and current API".format(legacy_key)
            description = "Record count is equal"
            _status = status.PASS
            reporter.addRow(title,description,_status,CURRENT_API=len(current_value), LEGACY_API=len(legacy_value))
            if legacy_value == current_value:
                title = f"Comparing | list of {legacy_key} for Legacy & Current API"
                description = "Values are equal"
                _status = status.PASS
                reporter.addRow(title,description,_status,CURRENT_API=current_info, LEGACY_API=legacy_info)
            else:
                title = f"Comparing | list of {legacy_key} for Legacy & Current API"
                description = "Values are not equal"
                _status = status.FAIL
                reporter.addRow(title,description,_status,CURRENT_API=current_info, LEGACY_API=legacy_info)
                reporter.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")
        elif len(legacy_value)==0 and len(current_value)==0:
            title = f"Comparing | list of {legacy_key} for Legacy & Current API"
            description = "Empty List"
            _status = status.PASS
            reporter.addRow(title,description,_status,CURRENT_API="[ ]", LEGACY_API="[ ]")
            

def responseReporter(reporter, legacy_key, current_key, legacy_value, current_value):
    title = ""
    description = ""
    _status = ""
    current_info = ""
    legacy_info = ""
    if legacy_key == current_key:
        title = f"Comparing | {legacy_key} for Legacy & Current API"
        current_info = f"value = {legacy_value}"
        legacy_info = f"Value = {current_value}"
        if legacy_value == current_value:
            description = f"Values are Equal"
            _status = status.PASS
            reporter.addRow(title, description, _status, CURRENT_API=current_info, LEGACY_API=legacy_info)
        else:
            description = f"Values are not Equal"
            _status = status.FAIL
            reporter.addRow(title, description, _status, CURRENT_API=current_info, LEGACY_API=legacy_info)
            reporter.addMisc("REASON OF FAILURE", "Mismatches found during Assertion")


def compareTypeOfResponses(legacy_response, current_response):
    logger.info("+++++++++++++++++++++ COMPARING TYPES OF RESPONSES +++++++++++++++++++++")
    if type(legacy_response) is type(current_response):
        logger.warning("+++++++++++++++++++++ TYPE MATCHED +++++++++++++++++++++")
        logger.info("+++++++++++++++++++++ Type of Legacy Response - {}".format(type(legacy_response).__name__))
        logger.info("+++++++++++++++++++++ Type of Current Response - {}".format(type(current_response).__name__))
        return True
    else:
        logger.warning("+++++++++++++++++++++ TYPE MISMATCHED +++++++++++++++++++++")
        logger.info("+++++++++++++++++++++ Type of Legacy Response - {}".format(type(legacy_response).__name__))
        logger.info("+++++++++++++++++++++ Type of Current Response - {}".format(type(current_response).__name__))
        return False


def compareNumberOfKeys(legacy_response_keys, current_response_keys):
    logger.info("+++++++++++++++++++++ COMPARING NUMBER OF KEYS OF RESPONSES +++++++++++++++++++++")
    if len(legacy_response_keys) == len(current_response_keys):
        logger.warning("+++++++++++++++++++++ Both responses have equal number of keys +++++++++++++++++++++")
        return True
    else:
        logger.warning("+++++++++++++++++++++ Both responses not have equal number of keys +++++++++++++++++++++")
        return False


def compareKeys(legacy_response_keys, current_response_keys):
    if legacy_response_keys == current_response_keys:
        logger.warning("+++++++++++++++++++++ Both responses have same keys +++++++++++++++++++++")
        return True
    else:
        logger.warning("+++++++++++++++++++++ Both responses do not have same keys +++++++++++++++++++++")
        return False


def compareStatusCodes(legacy_status_code, current_status_code):
    logger.info("+++++++++++++++++++++ COMPARING THE STATUS CODES OF RESPONSES +++++++++++++++++++++")
    if legacy_status_code == current_status_code:
        logger.warning("status codes are same %s, %s ",legacy_status_code, current_status_code)
        return True
    else:
        logger.warning("status codes are different %s, %s ",legacy_status_code, current_status_code)
        return False


from gempyp.config import DefaultSettings
import requests
import json
import logging
import traceback
from datetime import datetime


def closeJira(workflow):
    pass


def addComment(testcase_info, s_run_id, jira_id, email, access_token, jewel_link):
    comment_api = DefaultSettings.urls["data"]["comment-api"]
    comment_text = ""
    if testcase_info["PASS"] == testcase_info["TOTAL"]:
        comment_text = "SUITE PASSED \n"
    comment_text += "Date: {} \nS_RUN_ID: {} \nJewel Link: {} \n".format(datetime.now().strftime("%Y-%b-%d"), s_run_id, jewel_link)
    for statuses in testcase_info.keys():
        comment_text += statuses + ":" +  str(testcase_info.get(statuses)) + "\n"
    body = {"email": email, "accessToken": access_token, "jiraId": jira_id, "comment": comment_text}
    try:
        comment_res = requests.post(comment_api, data=json.dumps(body), headers={"Content-Type": "application/json"}, timeout=3)
        if comment_res.status_code == 201 or comment_res.status_code == 200:
            return jira_id
        else:
            logging.error("Comment could not be added")
    except Exception as e:
        logging.info(e)
        traceback.format_exc()
        return None


def createJira(email, access_token, title, project_id):
    logging.info("----------- Creating New Jira Ticket ------------")
    create_jira_api = DefaultSettings.urls["data"]["jira-api"]
    jira_body = {"email": email, "accessToken": access_token, "title": title, "projectId": project_id}
    try:
        jira_res = requests.post(create_jira_api, data=json.dumps(jira_body), headers={"Content-Type": "application/json"}, timeout=3)
        if jira_res.status_code == 201 or jira_res.status_code == 200:
            jira_json = json.loads(jira_res.text)
            jira_id = jira_json["data"]["key"]
            return jira_id
        else:
            logging.error("Unable to create the Jira")
    except Exception as e:
        traceback.format_exc()
        logging.info(e)
        return None


def jiraIntegration(s_run_id, suite_status, testcase_info, jewel_link, email, access_token, project, project_id, title=None, workflow=None):
    logging.info("---------- In Jira Integration ---------")
    if title is None:
        title = project
    api = DefaultSettings.urls["data"]["last-five"]
    params = {"s_run_id": s_run_id}
    jira_id = None
    try:
        response = requests.get(api, params=params, headers={"Content-Type": "application/json"}, timeout=3)
        prev_run_details = []
        if response.status_code == 200:
            prev_run_details = json.loads(response.text)
            jira_id = prev_run_details[0].get("Jira_id", None)
    except Exception as e:
        logging.info(e)
        return None
    if suite_status.upper() == "PASS" and prev_run_details[0]["Status"].upper() == "PASS":
        logging.info("----- \nCurrent suite_status is PASS and prev suite is also pass \nNo need to create jira \n------")
        return None
    elif suite_status.upper() == "PASS" and jira_id is not None and prev_run_details[0]["Status"].upper() not in ["PASS", "INFO"]:
        logging.info("----- Current suite_status is PASS ------")
        addComment(testcase_info, s_run_id, jira_id, email, access_token, jewel_link)
    elif suite_status.upper() == "FAIL" or suite_status.upper() == "ERR":
        try:
            if prev_run_details[0].get("Jira_id", None) is None:
                logging.info("------ Current suite_status is FAIL and no jira id for prev run ----")
                jira_id = createJira(email, access_token, title, project_id)
            elif prev_run_details[0]["Status"].upper() == "FAIL" or prev_run_details[0]["Status"].upper() == "ERR":
                logging.info("---------- Adding comment to the Jira Id {} -----------".format(prev_run_details[0]["Jira_id"]))
                jira_id = prev_run_details[0]["Jira_id"]
            elif prev_run_details[0]["Status"].upper() == "PASS" or prev_run_details[0]["Status"].upper() == "INFO":
                logging.info("----- Prev suite_status is PASS and current suite_status is FAIL or ERR -----")
                jira_id = createJira(email, access_token, title, project_id)
            jira_id = addComment(testcase_info, s_run_id, jira_id, email, access_token, jewel_link)
            return jira_id
        except Exception as e:
            logging.info(e)
            return None
        




import re
from gempyp.libs.enums.status import status
from gempyp.pyprest import utils
import logging as logging


class KeyCheck:
    def __init__(self, pyprest_obj):
        self.pyprest_obj = pyprest_obj
        self.regex_each = re.compile(r".*\[\beach\b\]")
        self.regex_int = re.compile(r".*\[\d+\]")
        self.isLegacyPresent = False
        pass

    def keyCheck(self):
        """
        Checks whether a key is present in the response or not.
        """
        self.pyprest_obj.logger.info("**************  INSIDE KEY CHECK  **************")

        self.response_body = {}

        if self.pyprest_obj.key_check and "keys are" in self.pyprest_obj.key_check.lower():
            self.pyprest_obj.key_check = self.pyprest_obj.key_check
            self.pyprest_obj.logger.info("keycheck string: " + str(self.pyprest_obj.key_check))
            type_list = self.pyprest_obj.key_check.strip(";").split(";")
            self.keys = []
            self.keys_not = []
            for each in type_list:
                # not condition
                if "keys are not" in each.lower():
                    key_not_str = each.split("keys are not")[1]
                    self.keys_not = [i.strip(" ") for i in key_not_str.split(",")]

                # keys are condition 
                elif "keys are" in each.lower():
                    key_str = each.split("keys are")[1]
                    keys = [i.strip(" ") for i in key_str.split(",")]
                    self.keys = list(set(keys) - set(self.keys_not))

                self.response_body = utils.formatRespBody(self.pyprest_obj.res_obj.response_body)
                if self.pyprest_obj.legacy_res is not None:
                    self.pyprest_obj.logger.info("Legacy API found for key check....")
                    self.isLegacyPresent = True
                    self.legacy_response_body = utils.formatRespBody(self.pyprest_obj.legacy_res.response_body)
                self.getKeys()

        else:
            self.pyprest_obj.logger.info("keycheck not performed----------------------")

    def getKeys(self):
        """
        Logging on the basis or presence or absence of keys
        """
        count_found_fail, count_notfound_fail = 0, 0
        _status, _status_n = status.INFO, status.INFO
        content_found = ""  # html content, add all the key checks in single row
        content_not_found = ""
        legacy_content_found = ""
        if len(self.keys) > 0:
            for each_key in self.keys:
                self.key_list = [str(i) for i in each_key.split(".")]
                self.temp_key_list = list(self.key_list)
                if self.isLegacyPresent:
                    result = ''
                    if 'legacy' in self.key_list[0]:
                        result = self.findKeys(self.legacy_response_body, self.key_list, self.temp_key_list,"legacy")
                        legacy_content_found += "" + ".".join(self.key_list) + "=" + result + " "

                    else:
                        result = self.findKeys(self.response_body, self.key_list,self.temp_key_list)
                        content_found += "" + ".".join(self.key_list) + "=" + result + " "
                    if "NOT FOUND" in result:
                        count_found_fail += 1
                    if count_found_fail > 0:
                        _status = status.FAIL
                    else:
                        _status = status.PASS
                    result = ''
                             
                else:
                    result = self.findKeys(self.response_body, self.key_list, self.temp_key_list)
                    if "NOT FOUND" in result:
                        count_found_fail += 1
                    content_found += "" + ".".join(self.key_list) + "=" + result + " "

                    if count_found_fail > 0:
                        _status = status.FAIL
                    else:
                        _status = status.PASS
            if self.isLegacyPresent:
                self.pyprest_obj.reporter.addRow("Keys to be check for presence/absence in response body of current and legacy api","Keys are present status",_status ,CURRENT_API= content_found, LEGACY_API= legacy_content_found)
                if _status == status.FAIL:
                    self.pyprest_obj.reporter.addMisc("REASON OF FAILURE", "Some keys are missing in Current or Legacy API Response")
            else:   
                self.pyprest_obj.reporter.addRow("Keys to be check for PRESENCE in response body", content_found, _status)
                if _status == status.FAIL:
                    self.pyprest_obj.reporter.addMisc("REASON OF FAILURE", "Some keys are missing in Response")

        if len(self.keys_not) > 0:
            for each_key in self.keys_not:
                self.key_list = [str(i) for i in each_key.split(".")]
                self.temp_key_list = list(self.key_list)
                if self.isLegacyPresent:
                    result = ''
                    if 'legacy' in self.key_list[0]:
                        result = self.findKeys(self.legacy_response_body, self.key_list, self.temp_key_list,"legacy")
                        legacy_content_found += "" + ".".join(self.key_list) + "=" + result + " "

                    else:
                        result = self.findKeys(self.response_body, self.key_list,self.temp_key_list)
                        content_found += "" + ".".join(self.key_list) + "=" + result + " "
                    if "NOT FOUND" in result:
                        count_found_fail += 1
                    if count_found_fail > 0:
                        _status = status.FAIL
                    else:
                        _status = status.PASS
                    result = ''
                else:
                    result = self.findKeys(self.response_body, self.key_list, self.temp_key_list)
                    if result == "FOUND":
                        count_notfound_fail += 1
                    content_not_found += "" + ".".join(self.key_list) + "=" + result + " "


                    if count_notfound_fail > 0:
                        _status_n = status.FAIL
                    else:
                        _status_n = status.PASS
            
            if self.isLegacyPresent:
                self.pyprest_obj.reporter.addRow("Keys to be check for presence/absence in response body of current and legacy api","Keys are not present in the response status",_status ,CURRENT_API= content_found, LEGACY_API= legacy_content_found)
                self.pyprest_obj.reporter.addMisc("REASON OF FAILURE", "Some keys are missing in Current or Legacy API Response")
            else:    
                self.pyprest_obj.reporter.addRow("Performing key check", "Checking Keys not required in response body - " + content_not_found, _status_n)

            if status.FAIL in [_status, _status_n]:
                self.pyprest_obj.reporter.addMisc("REASON OF FAILURE", "Status of key check is not as expected")
            # self.pyprest_obj.reporter.addMisc("REASON OF FAILURE", "Status of key check is not as expected")

# change adding another param typeOfResponse to make findkeys reusable for legacy and current responses

    def findKeys(self, json_data, key_list, temp_key_list, typeOfResponse = 'response'):
        """
        Finding keys in the response on the basis of the regex they match with
        """
        
        if "legacy" in key_list[0] or "legacy" in temp_key_list[0]:
            typeOfResponse = 'legacy'
        for each in key_list:
            self.pyprest_obj.logger.info(f"========FINDING KEY - {each}  ==========")
            """
            note : - if key to find is like data[0].color[9].code or something like data[each].color[9].code or any lind of nesting, then there will be recursion 
            and everytime the value after "." will be passed to the same function 
            i.e. first recusion --- for data[0]
                 second ----------- for color[9]
                 third ------------ for code
            """

            # parsing the keychecks with "each" in them
            if re.match(self.regex_each, each):
                br_index = each.find('[')
                key_val = each[:br_index]

                # for response[each].something
                if key_val.lower() == typeOfResponse and isinstance(json_data, list):
                    
                    key_list.remove(each)
                    return self.findKeys(json_data, key_list, temp_key_list)
                
                if key_val.lower() == typeOfResponse and isinstance(json_data, dict):
                    key_list.remove(each)

                    return self.findKeys(json_data, key_list, temp_key_list)

                # for data[each].something
                elif key_val in json_data and isinstance(json_data[key_val], list):
                    key_list.remove(each)
                    return self.findKeys(json_data[key_val], key_list, temp_key_list)

                # return "NOT FOUND" if  key from string "key[each]" is not found in the json_data
                else:
                    self.pyprest_obj.logger.info(f"{each}  - Key not found ------------")
                    return "NOT FOUND"

            # for list like keychecks
            # example  response[0].data, data[0].color etc

            elif re.match(self.regex_int, each):
                br_start = each.find('[')
                br_end = each.find(']')
                key_val = each[:br_start]
                key_num = int(each[br_start + 1:br_end])
                key_num, json_data = utils.getNestedListData(each, json_data, key_val)

                # for response[0].something
                if key_val == typeOfResponse and isinstance(json_data, list) and key_num < len(json_data):
                    key_list.remove(each)
                    return self.findKeys(json_data[key_num], key_list, temp_key_list)

                # for something like data[0].something
                elif key_val in json_data and isinstance(json_data[key_val], list) and key_num < len(json_data[key_val]):
                    key_list.remove(each)
                    return self.findKeys(json_data[key_val][key_num], key_list, temp_key_list)
                else:
                    self.pyprest_obj.logger.info(f"{each}  - Key not found ------------")
                    return "NOT FOUND"

            # for simple keychecks ( list and dicts) or last steps of list and "each" parsing
            else:

                # if input is list
                if isinstance(json_data, list):
                    temp_list = list(key_list)
                    i = temp_key_list.index(temp_list[0])
                    count_not_found = 0

                    # check if the parent key of this key_to_ckeck contained each or not, because other cases are already covered 
                    # example data[each].key_to_find
                    if json_data and re.match(self.regex_each, temp_key_list[i - 1]):
                        for name in json_data:
                            if isinstance(name, dict):
                                found_res = self.findKeys(name, temp_list, temp_key_list)
                                if found_res != "FOUND":
                                    count_not_found += 1
                            else:
                                return "NOT FOUND"
                        if count_not_found == 0:
                            return "FOUND"
                        else:

                            # return NOT FOUND if key not found in some_key[each][index]
                            return f"NOT FOUND, KEY MISSING IN {count_not_found} Items" # missing in number of objects( in case of each)

                    # return NOT FOUND if parent of key is not like some_key[each][index], because other scenarios of list are already covered
                    else:
                        return "NOT FOUND"

                # if input is dict
                elif isinstance(json_data, dict):
                    if each in json_data or each.lower() == typeOfResponse:
                        tmp_key = list(key_list)
                        tmp_key.remove(each)
                        if len(tmp_key) == 0:
                            return "FOUND"
                        if each.lower() == typeOfResponse:
                            json_data = json_data
                        else:
                            json_data = json_data[each]
                        return self.findKeys(json_data, tmp_key, temp_key_list)

                    # return not found if key not found in the dictionary i.e. the json_data . 
                    # json_data here is the json/list passed to this function in this recursion
                    else:
                        return "NOT FOUND"
                
                # return "NOT FOUND" if key doen not fall in any of the categories/conditions i.e. response[each] or some_key[each] or response[index] or some_key[index] or direct dictionary.key dict or li
                else:
                    return "NOT FOUND"

        # if the key w
        # as not found, then "NOT FOUND" must have been returned in any of the above conditions
        return "FOUND"



import traceback

class Database():

    def __init__(self, db_type, config_data) -> None:
        try:
            if db_type.lower() == "oracle":
                import data_compare.data.db_oracle as db
            elif db_type.lower() == 'mysql':
                import data_compare.data.db_mysql as db
            elif db_type.lower() == 'sqlite':
                import data_compare.data.db_sqlite as db
            elif db_type.lower() == 'postgre':
                import data.db_postgre as db   
            else:
                print("db_type not found")
                self.db_obj = None
            self.db_obj = db.Database(config_data)
        except Exception:
            self.db_obj = None
            print("Exception in db_wrapper {}".format(traceback.format_exc()))
    
    def get_db_obj(self):
        return self.db_obj



import importlib
class Databases:
    def __init__(self) -> None:
        pass

    def getConnectionString(self,db):
        database = {"snowflake": {"lib": "snowflake.connector", "conn": "connect"},
                    "mongoDB": {"lib": "pymongo", "conn": "MongoClient"},
                    "mysql":{"lib": "mysql.connector", "conn": "connect"},
                    "postgresql":{"lib": "pg8000", "conn": "connect"}}
        db_import = database[db]
        return db_import["lib"],db_import["conn"]
        



import os
import json
import logging
from datetime import datetime
import traceback
from typing import List, Union
import typing
from gempyp.config import DefaultSettings
import importlib
import sys
from gempyp.libs.gem_s3_common import download_from_s3, upload_to_s3, create_s3_link
from gempyp.config.GitLinkXML import fetchFileFromGit
from gempyp.config import DefaultSettings
from gempyp.engine import dataUpload
import logging
import requests
from gempyp.engine.dataUpload import _getHeaders
import re


def read_json(file_path):
    try:
        with open(file_path) as fobj:
            res = json.load(fobj)
    except Exception:
        res = None
    return res
def readPath(file_name):
        try:
            conf = file_name.split(os.sep)
            if conf[0] == ".." or conf[0]==".":
                script_path, script_name = importFromPath(file_name)
                for each in sys.path:
                    if isinstance(each,dict) and each is not None:
                        logging.info("Fetching config path - {}".format( each['XMLConfigDir']))
                        lib_path = os.path.join(each['XMLConfigDir'], file_name)
                        return lib_path
            else:
                return file_name
        except Exception as e:
            traceback.print_exc()


def findDuration(start_time: datetime, end_time: datetime):
    # finds the duration in the form HH MM SS

    duration = end_time - start_time
    seconds = duration.total_seconds()
    mins = seconds // 60
    seconds = seconds % 60

    if mins > 0:
        return f"{mins} mins and {round(seconds, 3)} seconds"
    return f"{round(seconds, 3)} seconds"


def errorHandler(logging, Error, msg="some Error Occured"):

    logging.error(msg)
    logging.error(f"Error: {Error}")

    if DefaultSettings.DEBUG:
        logging.error(f"TraceBack: {traceback.format_exc()}")


def parseMails(mail: Union[str, typing.TextIO]) -> List:
    try:
        if(mail is not None):
            if hasattr(mail, "read"):
                mails = mail.read()

            elif os.path.isfile(mail):
                file = open(mail, "r")
                mails = file.read()
                file.close()
            mails = mail.strip()
            emails=[]
            regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'
            for mail in mails.split(","):
                if(re.fullmatch(regex, mail)):
                    emails.append(mail)
            return emails
    except Exception as e:
        logging.error("Error while parsing the mails")
        logging.error(f"Error : {e}")
        logging.error(f"traceback: {traceback.format_exc()}")
        return []

def remove_none_from_dict(data):
    try:
        key_list = [key for key in data.keys()]
        for key in key_list:
            if isinstance(data[key], dict):
                remove_none_from_dict(data[key])
            if not data[key]:
                del data[key]
    except Exception as e:
        traceback.print_exc()
        pass
    return data


# custom encoder to encode date to epoch
class dateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.timestamp()*1000
        return json.JSONEncoder.default(self, o)

# absolute path path.... moved to file from runner due to circular import issue
# runner -> gempyp -> abstractsimpletestcase -> runner
def importFromPath(file_name):
    logging.info("--------- In import from path ----------")
    path_arr = file_name.split(os.sep)    
    file = path_arr[-1]
    path_arr.remove(file)
    path_cd = os.sep.join(path_arr)
    return path_cd, file
    
def moduleImports(file_names):
    for file_name in file_names:
        import_flag = 0
        if not file_name:
            return None
        try:
            logging.info("--------Trying importing modules--------")
            dynamicTestcase = importlib.import_module(file_name)       
            return dynamicTestcase
        except Exception as i:
            logging.info("-------Testcase not imported as module, Trying with absolute path-------")
            try:
                script_path, script_name = importFromPath(file_name)
                script_name = script_name.split(".")[0]
                if script_path is not None:
                    sys.path.append(script_path)
                try:
                    dynamicTestcase = importlib.import_module(script_name) 
                except Exception:
                    logging.info("when absolute and module import both failed. Trying relative path....")
                    try:
                        for each in sys.path:
                            if isinstance(each,dict) and each is not None:
                                logging.info("Fetching config path - {}".format( each['XMLConfigDir']))
                                lib_path = os.path.join(each['XMLConfigDir'], script_path)
                        sys.path.append(lib_path)
                        dynamicTestcase = importlib.import_module(script_name.split(".")[0])
                        import_flag = 1 
                    except Exception:
                        traceback.print_exc()
                    if import_flag != 1:
                        traceback.print_exc()
                return dynamicTestcase
            except Exception as e:
                logging.error("----- Error occured file could not be imported using any of the methods.-----")
                traceback.print_exc()
                return e

def download_common_file(file_names,headers=None):
     try:
        file_paths=[]
        file_names=file_names.split(",")
        for file_name in file_names: 
            if file_name and (file_name.__contains__('S3:')):
                logging.info("File is from S3")
                response=download_from_s3(api=file_name.replace("S3:",""),username=headers.get("username",None),bridge_token=headers.get("bridge_token",None))
                if(response.status_code==200):
                    file_name = os.path.join(file_name.split(":")[-1])
                    with open(file_name, "w+") as fp:
                        fp.seek(0)
                        fp.write(response.text)
                        fp.truncate()
                else:
                    logging.info(response.status_code)
                    logging.info(response.text)
            elif file_name and (file_name.__contains__('GIT')):
                logging.info("File is from GIT")
                list_url=file_name.split(":")
                if(len(list_url)>=5):
                    file_name=fetchFileFromGit(list_url[2],list_url[3],username=list_url[-2],bearer_token=list_url[-1])
                else:
                    file_name=fetchFileFromGit(list_url[2],list_url[3])
            file_paths.append(file_name)
        return file_paths
     except Exception as e:
        traceback.print_exc()
        return e
            

def control_text_size(data, **kwargs):

    fin_str = data
    if len(str(data)) > 150:
        url = None
        try:
            url = create_s3_link(url=upload_to_s3(DefaultSettings.urls["data"]["bucket-data-upload-api"], data=data, bridge_token=kwargs.get("bridge_token", None), username=kwargs.get("username", None), tag="public")["Url"])
            if url:
                fin_str = f'<a target=_blank  href={url}>Click here</a>'
        except Exception as e:
            logging.info(e)
    return fin_str


def check_json(data):
    try:
        data = json.loads(str(data))
    except Exception as e:
        logging.WARN(traceback.format_exc())

    return data

def get_reason_of_failure(data, e=None):  
    try:
        exceptiondata = data.splitlines()
        exceptionarray = [exceptiondata[-1]] + exceptiondata[1:-1]
        return exceptionarray[0]
    except:
        return e


def validateZeroTestcases(testcaseLength):
    if not testcaseLength:  # in case of zero testcases, we should not insert suite data 
            logging.warning("NO TESTCASES TO RUN..... PLEASE CHECK RUN FLAGS. ABORTING.................")
            sys.exit()
            


def runBaseUrls(jewel_user,base_url,username,bridgetoken):
    
            #trying rerun of base url api in case of api failure
            if base_url and DefaultSettings.apiSuccess == False:
                logging.info("Trying to call Api for getting urls")
                DefaultSettings.getEnterPoint(base_url ,bridgetoken,username)
            if not base_url:
                DefaultSettings.getEnterPoint(DefaultSettings.default_baseurl ,bridgetoken, username)
                logging.info("Base_url not found in suite data. Fetching default base_url")


def sendMail(s_run_id,mails,bridge_token,username):
    try:
        # payload={"to": mails, "s_run_id": s_run_id}
        mails["s_run_id"]=s_run_id
        response = requests.request(
        method="POST",
        url=DefaultSettings.getUrls("email-api"),
        data=json.dumps(mails),
        headers=_getHeaders(bridge_token, username),
        )
        if response.status_code == 200:
            logging.info("Report successfully sent on mail")
        else:
            logging.info(response.text)
    except Exception as e:
        traceback.print_exc()

def strValidation(string):
    return re.sub('[^A-Za-z0-9.-]+', '', string)


import tempfile
import traceback
from typing import Dict
import lxml.etree as et
import logging
from gempyp.config.baseConfig import AbstarctBaseConfig
from gempyp.libs.parsers import xmlToDict, xmlToList
from gempyp.libs.logConfig import LoggingConfig
import sys, os
import warnings
import uuid
from gempyp.config.customParser import CustomXMLParser
from lxml import etree
class XmlConfig(AbstarctBaseConfig):
    def __init__(self, filePath: str, s_run_id):
        
        # code pushed in parse function to make run_id working
        self.s_run_id = s_run_id
        super().__init__(filePath)
        # do any xml specific validatioins here

    def ignore_comments_and_parse(xml_file):
        parser = et.XMLParser(target=et.TreeBuilder(), comment_handler=lambda *args: None)
        tree = et.parse(xml_file, parser=parser)
        return tree


    def parse(self, file_paths):
        for filePath in file_paths:
            logging.info("-------- Xml file path: {filePath} ----------".format(filePath=filePath))
            path_list = filePath.split(os.sep)[0:-1]
            newfilePath = os.sep.join(path_list)
            sys.path.append({"XMLConfigDir":newfilePath})
            logging.info("-------- Started the Xml parsing in XmlConfig ---------")
            filePath=self.handleSpecialSymbols(filePath)
            # data = et.parse(filePath)
            parser = CustomXMLParser(remove_comments=True)
            data = etree.parse(filePath, parser=parser)
            if "SUITE_DATA" not in self._CONFIG:
                self._CONFIG["SUITE_DATA"] = self._getSuiteData(data) 
            #code for replacing variable from external properties file
            external_file_variables = None
            if self._CONFIG["SUITE_DATA"].get("PROPERTIES_FILE", None) is not None:
                try:
                    external_file_variables = self.read_variable_from_file(self._CONFIG["SUITE_DATA"]["PROPERTIES_FILE"])
                    if self._CONFIG["SUITE_DATA"].get("PROPERTIES_FILE"):
                            self._CONFIG["SUITE_DATA"] = self.replace_variables_from_file(external_file_variables,self._CONFIG["SUITE_DATA"])
                except Exception as e:
                        logging.info(traceback.print_exc())
            self.log_dir = str(os.path.join(tempfile.gettempdir(), 'logs'))
            if not os.path.exists(self.log_dir):
                os.makedirs(self.log_dir)
            self.unique_id = str(uuid.uuid4())
            if self.s_run_id != None:
                self.unique_id = self.s_run_id
            elif "S_RUN_ID" in self._CONFIG["SUITE_DATA"]:
                self.unique_id = self._CONFIG["SUITE_DATA"]["S_RUN_ID"]
            os.environ['unique_id'] = self.unique_id
            os.environ['log_dir'] = self.log_dir
            warnings.filterwarnings('ignore')
            suiteLogsLoc = str(os.path.join(self.log_dir, 'Suite_' + self.unique_id + '.txt'))  ## replacing log with txt for UI compatibility
            LoggingConfig(os.path.join(self.log_dir, 'Suite_' + self.unique_id + '.txt'))  ## replacing log with txt for UI compatibility
            # LoggingConfig(suiteLogsLoc)
            logging.info("suite logs : "+ str(suiteLogsLoc))
            current_testcase_data = self._getTestCaseData(data)
            if "TESTCASE_DATA" not in self._CONFIG:
                self._CONFIG["TESTCASE_DATA"] = current_testcase_data
            else:
            #     # Append the testcase data to the existing data
                self._CONFIG["TESTCASE_DATA"].update(current_testcase_data)
            if self._CONFIG["SUITE_DATA"].get("PROPERTIES_FILE") and external_file_variables:
                self.replace_variables_for_testcase(external_file_variables, self._CONFIG["TESTCASE_DATA"])
            self._CONFIG["SUITE_DATA"]['LOG_DIR'] = self.log_dir
            self._CONFIG["SUITE_DATA"]['UNIQUE_ID'] = self.unique_id
        return suiteLogsLoc

    def _getSuiteData(self, data) -> Dict:

        suite_data = data.find("suite")

        suite_dict = xmlToDict(suite_data)
        suite_dict={key.replace('-','_'):value for key,value in suite_dict.items()}
        suite_dict["SUITE_VARS"] = {}
        logging.info("suite_dict: \n {suite_dict} \n".format(suite_dict=suite_dict))
        # do your validations here

        return suite_dict

    def _getTestCaseData(self, data) -> Dict:

        testcase_data = data.find("testcases")

        testcase_list = xmlToList(testcase_data)

        for i in testcase_list:
            i["NAME"] = i["NAME"].upper()  ## uppercase

        testcase_list=[{key.replace('-','_'):value for key,value in testcase.items()} for testcase in testcase_list]

        testcase_dict = {k['NAME']: k for k in testcase_list}  #####################

        # do your validation here

        return testcase_dict
    
    def handleSpecialSymbols(self,filePath):
        filePath1=os.path.join(tempfile.gettempdir(),"temp.xml")
        f=open(filePath,"r")
        content=f.read()
        if(content.__contains__("&") and not(content.__contains__("&amp;"))):
            content=content.replace("&","&amp;")
        f1=open(filePath1,"w")
        f1.write(content)
        return filePath1
    
    def replace_substring(self,original_string, start_index, end_index, replacement):
        # Construct the modified string by concatenating the substring before and after the replacement
        modified_string = original_string[:start_index] + replacement + original_string[end_index + 1:]
        return modified_string

    def find_end_index_array(self,ch,string1)->list:
        """
        return list of indexes
        """
        pos= []
        for i in range(len(string1)):
            if ch == string1[i]:
                pos.append(i)
        return pos
    
    def read_variable_from_file(self, file_path)->Dict:
        print("Reading external file")
        files = file_path.split(',')
        final_variable_dict = {}
        for file in files:
            if os.path.exists(file):
                with open(file) as f:
                    l = [line.split("=") for line in f.readlines()]
                    d = {key.strip(): value.strip() for key, value in l}
                final_variable_dict = {**final_variable_dict, **d}
            else:
                print(f"File given in external files not found:{file}")
                sys.exit()
        return {key: (int(value) if value.isdigit() else float(value)) if value.replace('.', '', 1).isdigit() else value for key, value in final_variable_dict.items()}
        
    
    def replace_variables_from_file(self, variable_dict, suite_dict):
        logging.info("-- In external variable replacement--")
        values_with_variables = [[key,val] for key, val in suite_dict.items() if "$[#EXTERNAL." in val]
        for i in values_with_variables:
            string = i[1]
            start_indexes = [i for i in range(len(string)) if string.startswith("$[#EXTERNAL.", i)]
            for start_index in start_indexes:
                start_index = string.index("$[#EXTERNAL.")
                end_index = self.find_end_index_array("]",string)
                closest_value = min([i for i in end_index if i-start_index > 0])
                variable_name = string[start_index+12:closest_value]
                variable_value = variable_dict.get(variable_name,None)
                if variable_value == None:
                    logging.error("Value for variable {var} not found".format(var=variable_name))
                    sys.exit()
                if variable_value:
                    string = self.replace_substring(string, start_index,closest_value,variable_value)
            suite_dict[i[0]]= string
        return suite_dict


    def replace_variables_for_testcase(self,variable_dict,testcase_dict):
        try:
            for key,value in testcase_dict.items():
                self._CONFIG["TESTCASE_DATA"][key] = self.replace_variables_from_file(variable_dict,value)
        except Exception as e:
            logging.info(traceback.print_exc())






import enum


# IMP higher priority status must come above the lower priority in enum
# because the final status in report is based on the priority of the status
class status(enum.Enum):
    ERR = ("ERR",)
    FAIL = ("FAIL",)
    WARN = ("WARN",)
    PASS = ("PASS",)
    INFO = ("INFO",)
    EXE = "EXE"



import os
import sys
from datetime import date, datetime
import traceback
import decimal
# import cdecimal
import multiprocessing
# import Sybase
import pdb

DVM_NUMERIC_TYPES = [float, int, decimal.Decimal]
DVM_STR_TYPES = [str,bytearray]
DVM_DATETIME_TYPES = [datetime,datetime.time,datetime.date]
DVM_BOOL_TYPES = [bool]
#will add seconds and millis econs to this later
DEFAULT_DATETIME_FORMAT = "%Y-%m-%d %H:%M"

class ObjComparator():
    def __init__(self, src_data, tgt_data, keys, tolerance, src_header, tgt_header):
        self.src_data = src_data
        self.tgt_data = tgt_data
        print("Len of src_data {}".format(len(src_data)))
        print("Len of tgt_data {}".format(len(tgt_data)))
        print(src_data)
        print(tgt_data)
        self.keys = [x.lower() for x in keys]
        self.tolerance = tolerance
        self.src_header = src_header
        self.tgt_header = tgt_header
        #supposing the order of column in both src and target is same
        #to do - in pre processor, arranging the column from both src and tgt in same order without fetching data
        #then re running the query to create result set in ordered column fashion this is the case for select * query 
        #if user is providing the columns, they should order the columns similary on bothe src and tgt side
        self.src_key_idx = [idx for idx,item in enumerate(self.src_header) if item in self.keys]
        self.tgt_key_idx = [idx for idx,item in enumerate(self.tgt_header) if item in self.keys]
        self.src_val_idx = [idx for idx,item in enumerate(self.src_header) if item not in self.keys]
        self.tgt_val_idx = [idx for idx,item in enumerate(self.tgt_header) if item not in self.keys]
    
    def convert_to_dict(self):
        try:
            self.src_dict = {tuple([item[i] for i in self.src_key_idx ]):tuple([item[i] for i in self.src_val_idx]) for item in self.src_data }
            #need to add a check whether user inputted index column have duplicate value or not , i.e index value are really unique or not
            self.tgt_dict = {tuple([item[i] for i in self.tgt_key_idx ]):tuple([item[i] for i in self.tgt_val_idx]) for item in self.tgt_data }
        except Exception :
            print("Exception in convert to dict {}".format(traceback.format_exc()))

    def startCompare(self):
        try:
            r_val = 1
            print("Start compare at {}".format(datetime.now()))
            self.convert_to_dict()
            self.set_src_keys = set(self.src_dict.keys())
            self.set_tgt_keys = set(self.tgt_dict.keys())
            self.key_only_in_src = list(self.set_src_keys - self.set_tgt_keys)
            self.key_only_in_tgt = list(self.set_tgt_keys - self.set_src_keys)
            self.common_keys = list(self.set_tgt_keys.intersection(self.set_src_keys))
            print("key only in src {}".format(len(self.key_only_in_src)))
            print("key only in tgt {}".format(len(self.key_only_in_tgt)))
            print("common_keys {}".format(len(self.common_keys)))
            diff_count,match_count = compare(self.src_dict, self.tgt_dict, self.common_keys,self.tolerance)
            print("diffcount {}, match_count {}".format(diff_count,match_count))
            if diff_count == 0:
                r_val = 0
            else:
                r_val = diff_count
            res = (r_val, self.key_only_in_src, self.key_only_in_tgt, self.common_keys, diff_count, match_count)
            #print('res', res)
            print("Ended comparison at {}".format(datetime.now()))
        except Exception:
            print("Exception in startCompare {}".format(traceback.format_exc()))
            res = (1, None, None, None, None, None)
        return res


def isclose(a, b, rel_tol=1e-06, abs_tol = 0.0):
    """
    a and b : are the two values to be tested to relative closeness
    rel_tol: is the relatice tolerance -- it is the amount of error allowed, relative to the larger absolute value of a or b
    for example, to set a tolerance of 5%, pass tole=0.05 . the default tolerance is 1e-06
    whic assures that the two values are same within about 9 decimal digits. rel_tol must be greater than 0.0
    abs_tol = is a minimum absolute tolerance level--usefull for comaprison near zero
    """
    if type(rel_tol)!=float:
        try:
            rel_tol = float(rel_tol)
        except Exception:
            rel_tol = 1e-06
    return abs(a-b)<=max(rel_tol*max(abs(a),max(b),abs_tol))

def compare(srcdict, tgtdict, keys, tolerance, proc_queue=None, type_check = False, resType = 'basic'):
    """
    resType :(binary or basic or advance)
            binary result would be just pass/fail without diff inofrmation
            basic: result be put in json with diff information
            advance: complete report in excel
    """
    try:
        print(srcdict)
        print(tgtdict)
        match_count = 0
        diff_count = 0
        for key in keys:
            #matching the column at once, if they match we would do nothing, if they dont we will be matching column by column
            if srcdict[key] == tgtdict[key]:
                print("all values matched")
                match_count = match_count +len(srcdict[key])
            else:
                #now compare values one by one
                if len(srcdict[key]) == len(tgtdict[key]):
                    for val in range(len(srcdict[key])):
                        src_val = srcdict[key][val]
                        tgt_val = tgtdict[key][val]
                        src_type = type(src_val)
                        tgt_type = type(tgt_val)
                        #directly checking if the value matches or not
                        if src_val == tgt_val:
                            print("direct value matched")
                            match_count+=1
                            #val is same no diff
                        elif src_val is None or tgt_val is None:
                            diff_count+=1
                            #any one val is None so there is a diff
                        else:
                            #comparing string types
                            if src_type in DVM_STR_TYPES and tgt_type in DVM_STR_TYPES:
                                #str vs unicode is matched without issue
                                #str vs bytearray is matching without issue
                                #but bytearray vs unicode, we need to convert them to str
                                print("checking str type")
                                if src_type in [bytearray]:
                                    src_val = str(src_val)
                                if tgt_type in [bytearray]:
                                    tgt_val = str(tgt_val)
                                if src_val==tgt_val:
                                    match_count+=1
                                else:
                                    diff_count+=1
                            #matching numeric types:
                            elif src_type in DVM_NUMERIC_TYPES and tgt_type in DVM_NUMERIC_TYPES:
                                print("checking numeric type")
                                #only check for tolerance when type is either float or decimal.Decimal
                                if src_type in (float,decimal.Decimal) and tgt_type in (float,decimal.Decimal):
                                    #isclose only check toelrance for float value not for decimal as we will be converting every other type to float for tolerance comparisons
                                    if isclose(src_val,tgt_val,tolerance):
                                        match_count+=1
                                    else:
                                        diff_count+=1
                                else:
                                    #the number must be integer hcecking directly
                                    if src_val == tgt_val:
                                        match_count+=1
                                    else:
                                        diff_count+=1

                            #comparig datetime object
                            elif src_type in DVM_DATETIME_TYPES and tgt_type in DVM_DATETIME_TYPES:
                                print("checking datetime type")
                                if src_type ==datetime.date:
                                    src_val = datetime.combine(src_val,datetime.min.time())
                                if tgt_type==datetime.date:
                                    tgt_val = datetime.combine(tgt_val,datetime.min.time())
                                if src_val == tgt_val:
                                    match_count+=1
                                else:
                                    diff_count+=1
                            
                            #comparing bool type
                            elif src_type in DVM_BOOL_TYPES and tgt_type in DVM_BOOL_TYPES:
                                if src_val==tgt_val:
                                    match_count+=1
                                else:
                                    diff_count+=1
                            else:
                                diff_count+1
                                print("src {} and tgt type {} unidentified".format(src_type, tgt_type))
                
                if resType.lower() == 'binary':
                    #to do later
                    pass
  
        if proc_queue:
            proc_queue.put([diff_count, match_count])
        else:
            print("returning from try")
            return diff_count, match_count

    except Exception:
        print("Exception in compare method {}".format(traceback.format_exc()))
        if proc_queue:
            proc_queue.put([None, None])
        else:
            print("returnig from except")
            return None, None


if __name__ == "__main__":
    pass



from gempyp.config import DefaultSettings
import json
import logging
import traceback
from gempyp.engine import dataUpload


def createAzureTicket(body, bridge_token, user_name):
    logging.info("----------- Trying to create Azure Ticket ------------")
    create_azure_api = DefaultSettings.urls["data"]["azure-api"]

    try:

        azure_res = dataUpload._sendData(body, url=create_azure_api, bridge_token=bridge_token, user_name=user_name)
        azure_json = json.loads(azure_res.text)
        logging.info(azure_json)

        if(azure_json.get("operation").upper() == "FAILURE"):
            logging.warn(azure_json.get("message"))
        try:
            return azure_json.get("data", None).get("key", None)
        except:
            logging.info("No need to create azure")            
            return None
        
    except Exception as e:
        traceback.format_exc()
        logging.info(e)
        return None


def azureIntegration(s_run_id, assigned_to, azure_pat, fields, project_name, env, workflow, azureTestcaseFlag, title, bridge_token, user_name, suiteName, organization):  # adding title  ######################### post 1.0.4
    logging.info("---------- In Azure Integration ---------")
    
    azure_id = None
    if workflow:
        workflow = workflow.strip("'").strip('"').split(",")
        workflow = [str(i.strip(" ")) for i in workflow] 
   
    azure_body = {
        "assignedTo": assigned_to, 
        "pat": azure_pat,
        "project": project_name, 
        "organization": organization,
        "s_run_id": s_run_id,
        "flow": workflow,
        "env": env,
        "suiteName": suiteName,
        "title": title,
        "fields": fields,
        "azureTestcaseFlag" : azureTestcaseFlag
    }  
    if not assigned_to:
        del azure_body["assignedTo"]
    if not workflow:
        del azure_body["flow"]
    if not title:
        del azure_body["title"]  
    if not fields:
        del azure_body["fields"]
        
    azure_body = json.loads(json.dumps(str(azure_body).replace("'", '"')))
    try:
        logging.info("----------- Requesting AZURE API ------------")
        azure_id = createAzureTicket(azure_body, bridge_token, user_name)
        if azure_id:
            logging.info(f"Azure Id - {azure_id}")
        return azure_id
    except Exception as e:
        traceback.print_exc()
        return None
    




import logging as logger
import re
from inspect import getmembers, isfunction
import traceback
from gempyp.pyprest.predefinedFunctions import PredefinedFunctions as Prefunc
from gempyp.pyprest.variableReplacement import VariableReplacement as VarReplace



class PreVariables:
    def __init__(self, pyprest_obj):
        self.pyprest_obj = pyprest_obj
        functions_list = [x[0] for x in getmembers(Prefunc, isfunction)]
        self.functions_dict = {each.lower(): each for each in functions_list}
        pattern_func = r"\$\[\S[a-zA-Z0-9_.]+\(.*\)\]"
        self.regex_func = re.compile(pattern_func)
        # get variable written in data["PRE_VARIABLE"]
        # predefined func
        # remove $[#]
        # assign_values and append to a dict
        # different dicts for loal and suite variables
        


    def preVariable(self):
        """
        Maintaining 2 types of dictionaries- local and suile level"""
        if(len(self.pyprest_obj.list_subtestcases)==0):
            self.pyprest_obj.variables["local"] = {}
            # self.pyprest_obj.variables["suite"] = {}
            self.pyprest_obj.variables["suite"] = self.pyprest_obj.data.get("SUITE_VARS",{})
            self.pyprest_obj.variables["GLOBAL_VARIABLES"] = self.pyprest_obj.data.get("GLOBAL_VARIABLES", {})
        # print("self. suite vars in pre vars: ",self.pyprest_obj.data. )
        if self.pyprest_obj.pre_variables:
            self.pyprest_obj.logger.info("************** INSIDE PRE VARIABLES  **************")
            
            pre_variables_str = self.pyprest_obj.pre_variables

            # separate by ;
            variable_list = pre_variables_str.split(";")
            for each in variable_list:
                scope = 'local'
                each_item = each.split("=")
                if "SET" in each_item[0].upper() and '$[#' in each_item[0] and "GLOBAL" not in str(each_item[0]).upper():
                    # key = each_item[0].strip("set $[#").strip("Set $[#").strip("SET $[#").strip("]")
                    key = each_item[0].split("#")[1].replace("]","")
                    key = each_item[0].split("#")[1].replace("]","")
                    # find suite variables
                    if "SUITE." in key.upper():
                        scope = "suite"
                        key = key.replace(".", "_")
                    if "SUITE." in str(each_item[0].strip(" ")):
                        key = "SUITE_" + each_item[0].strip(" ").replace("set $[#SUITE.","").replace("]","").upper()
                        key = "SUITE_" + each_item[0].strip(" ").replace("set $[#SUITE.","").replace("]","").upper()
                        self.pyprest_obj.variables["suite"][key] = self.getFunctionValues(each_item[1])

                    self.pyprest_obj.variables[scope][key] = self.getFunctionValues(each_item[1])
                    key = "SUITE_" + each_item[0].strip(" ").strip("set $[#SUITE.").strip("]").upper()
                    self.pyprest_obj.variables["suite"][key] = self.getFunctionValues(each_item[1].strip().replace(" ",""))
                    self.pyprest_obj.variables[scope][key] = self.getFunctionValues(each_item[1].strip().replace(" ",""))
                
                    
            self.pyprest_obj.logger.info(f"Setting PRE VARIABLES: -------- {str(self.pyprest_obj.variables)}")

    def getFunctionValues(self, var_name):
        if re.match(self.regex_func, var_name):
            functions_list = [x[0] for x in getmembers(Prefunc, isfunction)]
            self.functions_dict = {each.lower():each for each in functions_list}
            func = var_name.strip("$[#]")
            func_items = func.split("(")
            func_name = func_items[0]
            params = func_items[1].strip(")")
            data = VarReplace(self.pyprest_obj).updateDataDictionary({"params": params})
            func_name = self.functions_dict.get(func_name.lower(), "invalid")

            try:
                Prefunc_ = Prefunc(self.pyprest_obj)
                params = Prefunc_.parseParams(data["params"])
                func_name = getattr(Prefunc_, func_name) if func_name != "invalid" else None
                if func_name is not None:
                    val =  func_name(*params)
                else:
                    val = "Null"
                return val
            except:
                self.pyprest_obj.logger.info(traceback.format_exc())
                return "Null"
        else:
            return var_name.strip(" ")